#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Integration Tests for `cellmaps_generate_hierarchy` package."""

import os

import unittest
from cellmaps_generate_hierarchy import cellmaps_generate_hierarchycmd

SKIP_REASON = 'CELLMAPS_GENERATE_HIERARCHY_INTEGRATION_TEST ' \
              'environment variable not set, cannot run integration ' \
              'tests'

@unittest.skipUnless(os.getenv('CELLMAPS_GENERATE_HIERARCHY_INTEGRATION_TEST') is not None, SKIP_REASON)
class TestIntegrationCellmaps_generate_hierarchy(unittest.TestCase):
    """Tests for `cellmaps_generate_hierarchy` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_something(self):
        """Tests parse arguments"""
        self.assertEqual(1, 1)



# -*- coding: utf-8 -*-

"""Top-level package for Cell Maps Generate Hierarchy."""

__author__ = 'Cell Maps team'
__email__ = 'tools@cm4ai.org'
__version__ = '0.2.1'
__repo_url__ = 'https://github.com/idekerlab/cellmaps_generate_hierarchy'
__description__ = 'A tool to generate hierarchies from protein to protein interaction networks'
__computation_name__ = 'Hierarchy'



import logging
from cellmaps_utils.ndexupload import NDExHierarchyUploader as NDExHierarchyUploaderUtil

logger = logging.getLogger(__name__)


class NDExHierarchyUploader(object):
    """
    Base class for uploading hierarchical networks and their parent networks to NDEx.

    Note:
        This class is deprecated and will be removed in a future release.
        Please use `NDExHierarchyUploader` from `cellmaps_utils.ndexupload` instead.

    """

    def __init__(self, ndexserver, ndexuser, ndexpassword, visibility=None):
        """
        Constructor

        :param ndexserver:
        :type ndexserver: str
        :param ndexuser:
        :type ndexuser: str
        :param ndexpassword:
        :type ndexpassword: str
        :param visibility: If set to ``public``, ``PUBLIC`` or ``True`` sets hierarchy and interactome to
                           publicly visibility on NDEx, otherwise they are left as private
        :type visibility: str or bool
        """
        self._ndex_uploader = NDExHierarchyUploaderUtil(ndexserver, ndexuser, ndexpassword, visibility)

    def get_cytoscape_url(self, ndexurl):
        """
        Generates a Cytoscape URL for a given NDEx network URL.

        :param ndexurl: The URL of the NDEx network.
        :type ndexurl: str
        :return: The URL pointing to the network's view on the Cytoscape platform.
        :rtype: str
        """
        return self._ndex_uploader.get_cytoscape_url(ndexurl)

    def save_hierarchy_and_parent_network(self, hierarchy, parent_ppi):
        """
        Saves both the hierarchy and its parent network to the NDEx server. This method first saves the parent
        network, then updates the hierarchy with HCX annotations based on the parent network's UUID, and
        finally saves the updated hierarchy. It returns the UUIDs and URLs for both the hierarchy and
        the parent network.

        :param hierarchy: The hierarchy network to be saved.
        :type hierarchy: :py:class:`~ndex2.cx2.CX2Network`
        :param parent_ppi: The parent protein-protein interaction network associated with the hierarchy.
        :type parent_ppi: :py:class:`~ndex2.cx2.CX2Network`
        :return: UUIDs and URLs for both the parent network and the hierarchy.
        :rtype: tuple
        """
        return self._ndex_uploader.save_hierarchy_and_parent_network(hierarchy, parent_ppi)

    def upload_hierary_and_parent_network_from_files(self, outdir):
        """
        Uploads hierarchy and parent network to NDEx from CX2 files located in a specified directory.
        It first checks the existence of the hierarchy and parent network files, then loads them into
        network objects, and finally saves them to NDEx using `save_hierarchy_and_parent_network` method.

        :param outdir: The directory where the hierarchy and parent network files are located.
        :type outdir: str
        :return: UUIDs and URLs for both the hierarchy and parent network.
        :rtype: tuple
        :raises CellmapsGenerateHierarchyError: If the required hierarchy or parent network files do not exist
                                                in the directory.
        """
        return self._ndex_uploader.upload_hierarchy_and_parent_network_from_files(outdir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `cellmaps_generate_hierarchy` package."""

import os
import tempfile
import shutil

import unittest
from cellmaps_generate_hierarchy import cellmaps_generate_hierarchycmd


class TestCellmaps_generate_hierarchy(unittest.TestCase):
    """Tests for `cellmaps_generate_hierarchy` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_parse_arguments(self):
        """Tests parse arguments"""
        res = cellmaps_generate_hierarchycmd._parse_arguments('hi',
                                                              ['outdir',
                                                               '--coembedding_dirs', 'foo',
                                                               '--ndexuser', 'foouser',
                                                               '--ndexpassword', 'foopass'])

        self.assertEqual('outdir', res.outdir)
        self.assertEqual(['foo'], res.coembedding_dirs)
        self.assertEqual(1, res.verbose)
        self.assertEqual(None, res.logconf)

        someargs = ['outdir', '-vv', '--logconf', 'hi',
                    '--coembedding_dirs', 'blah',
                    '--ndexuser', 'foouser',
                    '--ndexpassword', 'foopass']
        res = cellmaps_generate_hierarchycmd._parse_arguments('hi', someargs)

        self.assertEqual(3, res.verbose)
        self.assertEqual('hi', res.logconf)

    def test_main(self):
        """Tests main function"""

        # try where loading config is successful
        try:
            temp_dir = tempfile.mkdtemp()
            res = cellmaps_generate_hierarchycmd.main(['myprog.py',
                                                       temp_dir,
                                                       '--coembedding_dirs', 'foo',
                                                       '--ndexuser', 'foouser',
                                                       '--ndexpassword', 'foopass'])
            self.assertEqual(res, 2)
        finally:
            shutil.rmtree(temp_dir)



#! /usr/bin/env python

import os
import logging
import re
import time
import json
import warnings
from datetime import date

import ndex2
import pandas as pd
from tqdm import tqdm
from cellmaps_utils import constants
from cellmaps_utils import logutils
from cellmaps_utils.provenance import ProvenanceUtil
import cellmaps_generate_hierarchy
from cellmaps_generate_hierarchy.exceptions import CellmapsGenerateHierarchyError
from cellmaps_utils.hidefconverter import HierarchyToHiDeFConverter
from cellmaps_utils.ndexupload import NDExHierarchyUploader

from cellmaps_generate_hierarchy.hcx import HCXFromCDAPSCXHierarchy

logger = logging.getLogger(__name__)


class CellmapsGenerateHierarchy(object):
    """
    Runs steps necessary to create PPI from embedding and to
    generate a hierarchy
    """
    K_DEFAULT = 10
    ALGORITHM = 'leiden'
    MAXRES = 80

    def __init__(self, outdir=None,
                 inputdirs=[],
                 ppigen=None,
                 algorithm=ALGORITHM,
                 maxres=MAXRES,
                 k=K_DEFAULT,
                 gene_node_attributes=None,
                 hiergen=None,
                 name=None,
                 organization_name=None,
                 project_name=None,
                 layoutalgo=None,
                 skip_logging=True,
                 provenance_utils=ProvenanceUtil(),
                 input_data_dict=None,
                 ndexserver=None,
                 ndexuser=None,
                 ndexpassword=None,
                 visibility=None,
                 keep_intermediate_files=False,
                 provenance=None
                 ):
        """
        Constructor

        :param outdir: Directory to create and put results in
        :type outdir: str
        :param ppigen: PPI Network Generator object, should be a subclass
        :type ppigen: :py:class:`~cellmaps_generate_hierarchy.ppi.PPINetworkGenerator`
        :param hiergen: Hierarchy Generator object, should be a subclass
        :type hiergen: :py:class:`~cellmaps_generate_hierarchy.HierarchyGenerator`
        :param name:
        :param organization_name:
        :param project_name:
        :param skip_logging: If ``True`` skip logging, if ``None`` or ``False`` do NOT skip logging
        :type skip_logging: bool
        :param provenance_utils:
        :param ndexserver:
        :type ndexserver: str
        :param ndexuser:
        :type ndexuser: str
        :param ndexpassword:
        :type ndexpassword: str
        :param visibility: If set to ``public``, ``PUBLIC`` or ``True`` sets hierarchy and interactome to
                           publicly visibility on NDEx, otherwise they are left as private
        :type visibility: str or bool
        """
        logger.debug('In constructor')
        if outdir is None:
            raise CellmapsGenerateHierarchyError('outdir is None')
        self._outdir = os.path.abspath(outdir)
        self._inputdirs = inputdirs
        self._start_time = int(time.time())
        self._ppigen = ppigen
        self._algorithm = algorithm
        self._maxres = maxres
        self._k = k
        self._gene_node_attributes = gene_node_attributes
        self._hiergen = hiergen
        self._name = name
        self._project_name = project_name
        self._organization_name = organization_name
        self._keywords = None
        self._description = None
        if skip_logging is None:
            self._skip_logging = False
        else:
            self._skip_logging = skip_logging
        self._input_data_dict = input_data_dict
        self._provenance_utils = provenance_utils
        self._layoutalgo = layoutalgo
        self._server = ndexserver
        self._user = ndexuser
        self._password = ndexpassword
        self._visibility = visibility
        self.keep_intermediate_files = keep_intermediate_files
        self._provenance = provenance

        if self._input_data_dict is None:
            self._input_data_dict = {'outdir': self._outdir,
                                     'inputdirs': self._inputdirs,
                                     'embedding_generator': str(self._ppigen),
                                     'algorithm': self._algorithm,
                                     'maxres': self._maxres,
                                     'k': self._k,
                                     'gene_node_attributes': str(self._gene_node_attributes),
                                     'hiergen': str(self._hiergen),
                                     'ndexserver': self._server,
                                     'ndexuser': self._user,
                                     'ndexpassword': self._password,
                                     'name': self._name,
                                     'project_name': self._project_name,
                                     'organization_name': self._organization_name,
                                     'skip_logging': self._skip_logging,
                                     'provenance': str(self._provenance)
                                     }

    def _update_provenance_fields(self):
        """

        :return:
        """
        rocrate_dirs = []
        if self._inputdirs is not None:
            if isinstance(self._inputdirs, str):
                if os.path.exists(os.path.join(os.path.abspath(self._inputdirs), constants.RO_CRATE_METADATA_FILE)):
                    rocrate_dirs.append(self._inputdirs)
            else:
                for embeddind_dir in self._inputdirs:
                    if os.path.exists(os.path.join(os.path.abspath(embeddind_dir), constants.RO_CRATE_METADATA_FILE)):
                        rocrate_dirs.append(embeddind_dir)
        if len(rocrate_dirs) > 0:
            prov_attrs = self._provenance_utils.get_merged_rocrate_provenance_attrs(self._inputdirs,
                                                                                    override_name=self._name,
                                                                                    override_project_name=
                                                                                    self._project_name,
                                                                                    override_organization_name=
                                                                                    self._organization_name,
                                                                                    extra_keywords=['hierarchy',
                                                                                                    'model'])

            self._name = prov_attrs.get_name()
            self._organization_name = prov_attrs.get_organization_name()
            self._project_name = prov_attrs.get_project_name()
            self._keywords = prov_attrs.get_keywords()
            self._description = prov_attrs.get_description()
        elif self._provenance is not None:
            self._name = self._provenance['name'] if 'name' in self._provenance else 'Hierarchy'
            self._organization_name = self._provenance['organization-name'] \
                if 'organization-name' in self._provenance else 'NA'
            self._project_name = self._provenance['project-name'] \
                if 'project-name' in self._provenance else 'NA'
            self._keywords = self._provenance['keywords'] if 'keywords' in self._provenance else ['hierarchy', 'model']
            self._description = self._provenance['description'] if 'description' in self._provenance else \
                'Hierarchy generation'
        else:
            raise CellmapsGenerateHierarchyError('One of inputs directories should be an RO-Crate or provenance file '
                                                 'should be specified.')

    def _create_rocrate(self):
        """
        Creates rocrate for output directory

        :raises CellMapsProvenanceError: If there is an error
        """
        logger.debug('Registering rocrate with FAIRSCAPE')

        try:
            self._provenance_utils.register_rocrate(self._outdir,
                                                    name=self._name,
                                                    organization_name=self._organization_name,
                                                    project_name=self._project_name,
                                                    description=self._description,
                                                    keywords=self._keywords)
        except TypeError as te:
            raise CellmapsGenerateHierarchyError('Invalid provenance: ' + str(te))
        except KeyError as ke:
            raise CellmapsGenerateHierarchyError('Key missing in provenance: ' + str(ke))

    def _get_keywords_extended_with_new_values(self, new_values=None):
        """
        Takes keywords passed into constructor and append **new_values**
        and return a unique list of merged values

        :param new_values: new values
        :type new_values: list
        :return: merged list of keywords
        :rtype: list
        """
        if self._keywords is None or len(self._keywords) == 0:
            keywords = []
        else:
            keywords = self._keywords.copy()

        if isinstance(new_values, list):
            keywords.extend(new_values)
        else:
            keywords.append(new_values)

        return list(set(keywords))

    def _register_software(self):
        """
        Registers this tool

        :raises CellMapsImageEmbeddingError: If fairscape call fails
        """
        software_keywords = self._get_keywords_extended_with_new_values(new_values=['tools',
                                                                                    cellmaps_generate_hierarchy.__name__])
        software_description = self._description + ' ' + \
                               cellmaps_generate_hierarchy.__description__
        self._softwareid = self._provenance_utils.register_software(self._outdir,
                                                                    name=cellmaps_generate_hierarchy.__name__,
                                                                    description=software_description,
                                                                    author=cellmaps_generate_hierarchy.__author__,
                                                                    version=cellmaps_generate_hierarchy.__version__,
                                                                    file_format='py',
                                                                    keywords=software_keywords,
                                                                    url=cellmaps_generate_hierarchy.__repo_url__)

    def _register_computation(self, generated_dataset_ids=[]):
        """
        # Todo: added in used dataset, software and what is being generated
        :return:
        """
        logger.debug('Getting id of input rocrate')
        input_dataset_ids = []
        if isinstance(self._inputdirs, list):
            for i_dir in self._inputdirs:
                input_dataset_ids.append(self._provenance_utils.get_id_of_rocrate(i_dir))
        else:
            input_dataset_ids.append(self._provenance_utils.get_id_of_rocrate(self._inputdirs))

        keywords = self._get_keywords_extended_with_new_values(new_values=['computation'])
        description = self._description + ' run of ' + cellmaps_generate_hierarchy.__name__
        self._provenance_utils.register_computation(self._outdir,
                                                    name=cellmaps_generate_hierarchy.__computation_name__,
                                                    run_by=str(self._provenance_utils.get_login()),
                                                    command=str(self._input_data_dict),
                                                    description=description,
                                                    keywords=keywords,
                                                    used_software=[self._softwareid],
                                                    used_dataset=input_dataset_ids,
                                                    generated=generated_dataset_ids)

    def get_ppi_network_dest_file(self, ppi_network):
        """
        Gets the path where the PPI network should be written to

        :param ppi_network: PPI Network
        :type ppi_network: :py:class:`ndex2.nice_cx_network.NiceCXNetwork`
        :return: Path on filesystem to write the PPI network
        :rtype: str
        """
        cutoff = ppi_network.get_network_attribute('cutoff')['v']
        return os.path.join(self._outdir, constants.PPI_NETWORK_PREFIX +
                            '_cutoff_' + str(cutoff))

    def get_hierarchy_dest_file(self):
        """
        Creates file path prefix for hierarchy

        Example path: ``/tmp/foo/hierarchy``

        :return: Prefix path on filesystem to write Hierarchy Network
        :rtype: str
        """
        return os.path.join(self._outdir, constants.HIERARCHY_NETWORK_PREFIX)

    def get_hierarchy_parent_network_dest_file(self):
        """
        Creates file path prefix for hierarchy parent network

        Example path: ``/tmp/foo/hierarchy_parent``
        :return:
        """
        return os.path.join(self._outdir, 'hierarchy_parent')

    def _remove_ppi_networks(self, networks_paths):
        for n in networks_paths:
            try:
                os.remove(n + constants.CX_SUFFIX)
            except Exception as e:
                logger.warning(f"Tried to remove ppi file {n}, but failed due to: {e}")

    def _write_ppi_network_as_cx(self, ppi_network, dest_path=None):
        """

        :param ppi_network:
        :return:
        """
        logger.debug('Writing PPI network ' + str(ppi_network.get_name()))
        # write PPI to filesystem

        with open(dest_path, 'w') as f:
            json.dump(ppi_network.to_cx(), f)

    def _register_ppi_network(self, ppi_network, dest_path=None):
        """

        :param ppi_network:
        :return:
        """
        logger.debug('Registering PPI network ' + str(ppi_network.get_name()))

        description = self._description
        description += ' PPI Network file'

        keywords = self._get_keywords_extended_with_new_values(new_values=['file'])

        # register ppi network file with fairscape
        data_dict = {'name': os.path.basename(dest_path) + ' PPI network file',
                     'description': description,
                     'keywords': keywords,
                     'data-format': 'CX',
                     'author': cellmaps_generate_hierarchy.__name__,
                     'version': cellmaps_generate_hierarchy.__version__,
                     'date-published': date.today().strftime(self._provenance_utils.get_default_date_format_str())}
        return self._provenance_utils.register_dataset(self._outdir,
                                                       source_file=dest_path,
                                                       data_dict=data_dict)

    def _write_hierarchy_network(self, hierarchy=None):
        """
        Writes **hierarchy** to file

        :param hierarchy: CX2 network converted to list and dicts
        :type hierarchy: list
        :return: Path to hierarchy output file
        :rtype: str
        """
        logger.debug('Writing hierarchy')
        suffix = '.cx2'  # todo put this into cellmaps_utils.constants
        hierarchy_out_file = self.get_hierarchy_dest_file() + suffix
        with open(hierarchy_out_file, 'w') as f:
            json.dump(hierarchy, f)

        return hierarchy_out_file

    def _register_hierarchy_network(self, hierarchy_out_file=None, hierarchyurl=None):
        """

        :param network:
        :return:
        """
        logger.debug('Register hierarchy with fairscape')

        description = self._description
        description += ' Hierarchy network file'
        keywords = self._get_keywords_extended_with_new_values(new_values=['file',
                                                                           'hierarchy',
                                                                           'network',
                                                                           'HCX'])
        # register hierarchy network file with fairscape
        # The name must be Output Dataset so that the cm4ai portal knows to
        # grab the URL link
        data_dict = {'name': 'Output Dataset',
                     'description': description,
                     'keywords': keywords,
                     'data-format': 'HCX',
                     'author': cellmaps_generate_hierarchy.__name__,
                     'version': cellmaps_generate_hierarchy.__version__,
                     'date-published': date.today().strftime(self._provenance_utils.get_default_date_format_str())}
        if hierarchyurl is not None:
            data_dict['url'] = hierarchyurl
        dataset_id = self._provenance_utils.register_dataset(self._outdir,
                                                             source_file=hierarchy_out_file,
                                                             data_dict=data_dict)
        return dataset_id

    def _write_and_register_hierarchy_parent_network(self, parent=None, parenturl=None):
        """

        :param network:
        :return:
        """
        logger.debug('Writing hierarchy parent')
        suffix = '.cx2'  # todo put this into cellmaps_utils.constants
        parent_out_file = self.get_hierarchy_parent_network_dest_file() + suffix
        with open(parent_out_file, 'w') as f:
            json.dump(parent, f)
        description = self._description
        description += ' Hierarchy parent network file'
        keywords = self._get_keywords_extended_with_new_values(new_values=['file',
                                                                           'parent',
                                                                           'interactome',
                                                                           'ppi',
                                                                           'network',
                                                                           'CX2'])

        data_dict = {'name': 'Hierarchy parent network',
                     'description': description,
                     'keywords': keywords,
                     'data-format': 'CX2',
                     'author': cellmaps_generate_hierarchy.__name__,
                     'version': cellmaps_generate_hierarchy.__version__,
                     'date-published': date.today().strftime(self._provenance_utils.get_default_date_format_str())}
        if parenturl is not None:
            data_dict['url'] = parenturl
        dataset_id = self._provenance_utils.register_dataset(self._outdir,
                                                             source_file=parent_out_file,
                                                             data_dict=data_dict)
        return dataset_id

    def _register_hidef_output_with_gene_names(self, hidef_output_path, hidef_output_name):
        """
        """
        logger.debug(f'Registering hidef output {hidef_output_name} with gene names')

        description = self._description
        description += f' HiDeF output {hidef_output_name} with gene names file'

        keywords = self._get_keywords_extended_with_new_values(new_values=['file'])

        # register file with fairscape
        data_dict = {'name': f'HiDeF output {hidef_output_name} with gene names',
                     'description': description,
                     'keywords': keywords,
                     'data-format': "tsv",
                     'author': cellmaps_generate_hierarchy.__name__,
                     'version': cellmaps_generate_hierarchy.__version__,
                     'date-published': date.today().strftime(self._provenance_utils.get_default_date_format_str())}
        return self._provenance_utils.register_dataset(self._outdir,
                                                       source_file=hidef_output_path,
                                                       data_dict=data_dict)

    def _add_gene_node_attributes(self, parent_ppi):
        """
        Adds gene node attributes to the parent PPI network from provided TSV files or found in ro-crates.

        :param parent_ppi: The PPI network to which the attributes will be added.
        :type parent_ppi: :py:class:`ndex2.cx2.CX2Network`
        :return: The parent PPI network object with the new attributes added.
        :rtype: :py:class:`ndex2.cx2.CX2Network`
        """
        node_name_dict = {}
        for node_id, node_obj in parent_ppi.get_nodes().items():
            node_name_dict[node_obj['v']['name']] = node_id

        for entry_path in self._gene_node_attributes:
            attr_files = list()
            if os.path.isdir(entry_path):
                attr_files.extend([os.path.join(entry_path, f) for f in os.listdir(entry_path)
                                   if re.match(r'\d+_' + re.escape(constants.IMAGE_GENE_NODE_ATTR_FILE), f)])

                ppi_attr_file = os.path.join(entry_path, constants.PPI_GENE_NODE_ATTR_FILE)
                if os.path.exists(ppi_attr_file):
                    attr_files.append(ppi_attr_file)

                if len(attr_files) < 1:
                    logger.warning(f"No attribute file found in directory {entry_path}")
                    continue
            elif entry_path.endswith('.tsv'):
                attr_files.append(entry_path)
            else:
                logger.warning(f"Entry is neither a directory nor a TSV file: {entry_path}")
                continue

            for attribute_file in attr_files:
                df = pd.read_csv(attribute_file, sep='\t', header=0)

                for _, row in df.iterrows():
                    gene_name = row.iloc[0]
                    node_id = node_name_dict.get(gene_name, None)
                    if node_id is None:
                        continue

                    for column_name in df.columns[1:]:
                        if not pd.isna(row[column_name]):
                            parent_ppi.add_node_attribute(node_id, column_name, row[column_name])
                            if column_name == 'represents' and row[column_name].startswith('ensembl:ENSG'):
                                ensembl_only = re.sub('^ensembl:', '', row[column_name])
                                # URL suggested by Jan to get HPA info
                                parent_ppi.add_node_attribute(node_id, 'representsurl',
                                                              'https://www.proteinatlas.org/' +
                                                              ensembl_only + '/subcellular')

                                # URL suggested by Jan to get all antibodies for given ensembl id
                                parent_ppi.add_node_attribute(node_id, 'antibodyurl',
                                                              'https://www.proteinatlas.org/' +
                                                              ensembl_only + '/summary/antibody')
        return parent_ppi

    def _get_network_attribute(self, network=None, attribute_name=None,
                               default='Unknown'):
        """
        Gets network attribute from **network** value matching
        **attribute_name** or value of **default** if not found

        :param network:
        :type network: :py:class:`~ndex2.cx2.CX2Network`
        :param attribute_name:
        :type attribute_name: str
        :param default:
        :type default: str
        :return:
        :rtype: str
        """
        net_attrs = network.get_network_attributes()
        if net_attrs is None:
            logger.info('Network lacks any network attributes. hmm....')
            return default
        if attribute_name in net_attrs:
            return net_attrs[attribute_name]
        logger.debug(str(attribute_name) + ' network attribute note found. using default')
        return default

    def _update_ppi_with_hierarchy_attributes(self, parent_ppi=None, hierarchy=None):
        """
        Updates parent_ppi aka parent network with some attributes from hierarchy
        namely **prov:wasGeneratedBy** and **prov:wasDerivedFrom**

        In addition
        :param hierarchy:
        :type hierarchy: :py:class:`~ndex2.cx2.CX2Network`
        """
        parent_ppi.add_network_attribute('prov:wasGeneratedBy',
                                         self._get_network_attribute(hierarchy, attribute_name='prov:wasGeneratedBy'))
        parent_ppi.add_network_attribute('prov:wasDerivedFrom',
                                         self._get_network_attribute(hierarchy, attribute_name='prov:wasDerivedFrom'))
        p_net_attrs = parent_ppi.get_network_attributes()
        parent_ppi.add_network_attribute('name', str(hierarchy.get_name() + ' ' + p_net_attrs['name']))

    def generate_readme(self):
        description = getattr(cellmaps_generate_hierarchy, '__description__', 'No description provided.')
        version = getattr(cellmaps_generate_hierarchy, '__version__', '0.0.0')

        with open(os.path.join(os.path.dirname(__file__), 'readme_outputs.txt'), 'r') as f:
            readme_outputs = f.read()

        readme = readme_outputs.format(DESCRIPTION=description, VERSION=version)
        with open(os.path.join(self._outdir, 'README.txt'), 'w') as f:
            f.write(readme)

    def run(self):
        """
        Runs CM4AI Generate Hierarchy


        :return:
        """
        exitcode = 99
        try:
            logger.debug('In run method')

            if os.path.isdir(self._outdir):
                raise CellmapsGenerateHierarchyError(self._outdir + ' already exists')
            if not os.path.isdir(self._outdir):
                os.makedirs(self._outdir, mode=0o755)
            if self._skip_logging is False:
                logutils.setup_filelogger(outdir=self._outdir,
                                          handlerprefix='cellmaps_image_embedding')
            logutils.write_task_start_json(outdir=self._outdir,
                                           start_time=self._start_time,
                                           data={'commandlineargs': self._input_data_dict},
                                           version=cellmaps_generate_hierarchy.__version__)

            self.generate_readme()

            self._update_provenance_fields()

            self._create_rocrate()

            self._register_software()

            generated_dataset_ids = []
            ppi_network_prefix_paths = []
            # generate PPI networks
            for ppi_network in tqdm(self._ppigen.get_next_network(), desc='Generating hierarchy'):
                dest_prefix = self.get_ppi_network_dest_file(ppi_network)
                ppi_network_prefix_paths.append(dest_prefix)
                cx_path = dest_prefix + constants.CX_SUFFIX
                self._write_ppi_network_as_cx(ppi_network, dest_path=cx_path)
                if self.keep_intermediate_files:
                    generated_dataset_ids.append(self._register_ppi_network(ppi_network, dest_path=cx_path))

            # generate hierarchy and get parent ppi
            hierarchy, parent_ppi = self._hiergen.get_hierarchy(ppi_network_prefix_paths, self._algorithm, self._maxres,
                                                                self._k)

            if not self.keep_intermediate_files:
                self._remove_ppi_networks(ppi_network_prefix_paths)

            if self._gene_node_attributes is not None:
                parent_ppi = self._add_gene_node_attributes(parent_ppi)
                if "bait" in parent_ppi.get_attribute_declarations()['nodes']:
                    parent_ppi = HCXFromCDAPSCXHierarchy.apply_style_to_network(parent_ppi,
                                                                                'interactome_style_with_bait.cx2')

            parenturl = None
            hierarchyurl = None

            self._update_ppi_with_hierarchy_attributes(parent_ppi=parent_ppi, hierarchy=hierarchy)

            if self._server is not None and self._user is not None and self._password is not None:
                ndex_uploader = NDExHierarchyUploader(self._server, self._user, self._password, self._visibility)
                _, parenturl, _, hierarchyurl = ndex_uploader.save_hierarchy_and_parent_network(hierarchy, parent_ppi)
                message = (f'Hierarchy uploaded. To view hierarchy on NDEx please paste this URL in your browser '
                           f'{hierarchyurl}. To view Hierarchy on new experimental Cytoscape on the Web, '
                           f'go to {ndex_uploader.get_cytoscape_url(hierarchyurl)}')
                print(message)
                logger.info(message)

            hierarchy = hierarchy.to_cx2()
            parent_ppi = parent_ppi.to_cx2()

            # TODO: Need to support layout with HCX
            warnings.warn("Layout disabled due to incompatibilities with HCX format")
            # if self._layoutalgo is not None:
            #    logger.debug('Applying layout')
            #    self._layoutalgo.add_layout(network=hierarchy)
            # else:
            #    logger.debug('No layout algorithm set, skipping')

            # write out hierarchy
            hierarchy_out_file = self._write_hierarchy_network(hierarchy)

            # write out parent network and register with fairscape
            generated_dataset_ids.append(self._write_and_register_hierarchy_parent_network(parent=parent_ppi,
                                                                                           parenturl=parenturl))

            generated_dataset_ids.append(self._register_hierarchy_network(hierarchy_out_file,
                                                                          hierarchyurl=hierarchyurl))

            # add datasets created by hiergen object
            generated_dataset_ids.extend(self._hiergen.get_generated_dataset_ids())

            hidef_converter = HierarchyToHiDeFConverter(self._outdir, self._outdir)
            hidef_nodes, hidef_edges = hidef_converter.generate_hidef_files()
            generated_dataset_ids.append(
                self._register_hidef_output_with_gene_names(hidef_nodes, 'nodes'))
            generated_dataset_ids.append(
                self._register_hidef_output_with_gene_names(hidef_edges, 'edges'))

            # register generated datasets
            self._register_computation(generated_dataset_ids=generated_dataset_ids)
            exitcode = 0
        finally:
            logutils.write_task_finish_json(outdir=self._outdir,
                                            start_time=self._start_time,
                                            status=exitcode)

        return exitcode



#! /usr/bin/env python

import argparse
import json
import os
import sys
import logging
import logging.config
import getpass

from cellmaps_utils import logutils
from cellmaps_utils import constants
from cellmaps_utils.provenance import ProvenanceUtil
import cellmaps_generate_hierarchy
from cellmaps_generate_hierarchy.exceptions import CellmapsGenerateHierarchyError
from cellmaps_utils.hidefconverter import HierarchyToHiDeFConverter
from cellmaps_utils.ndexupload import NDExHierarchyUploader
from cellmaps_generate_hierarchy.ppi import CosineSimilarityPPIGenerator
from cellmaps_generate_hierarchy.hierarchy import CDAPSHiDeFHierarchyGenerator
from cellmaps_generate_hierarchy.maturehierarchy import HiDeFHierarchyRefiner
from cellmaps_generate_hierarchy.runner import CellmapsGenerateHierarchy
from cellmaps_generate_hierarchy.layout import CytoscapeJSBreadthFirstLayout
from cellmaps_generate_hierarchy.hcx import HCXFromCDAPSCXHierarchy

logger = logging.getLogger(__name__)

CO_EMBEDDINGDIRS = '--coembedding_dirs'


def _parse_arguments(desc, args):
    """
    Parses command line arguments

    :param desc: description to display on command line
    :type desc: str
    :param args: command line arguments usually :py:func:`sys.argv[1:]`
    :type args: list
    :return: arguments parsed by :py:mod:`argparse`
    :rtype: :py:class:`argparse.Namespace`
    """
    parser = argparse.ArgumentParser(description=desc,
                                     formatter_class=constants.ArgParseFormatter)
    parser.add_argument('outdir', help='Output directory')
    parser.add_argument(CO_EMBEDDINGDIRS, nargs="+",
                        help='Directories where coembedding was run')
    parser.add_argument('--mode', choices=['run', 'ndexsave', 'convert'], default='run',
                        help='Processing mode. If set to "run" then hierarchy is generated. If '
                             'set to "ndexsave", it is assumes hierarchy has been generated '
                             '(named hierarchy.cx2 and parent_hierarchy.cx2) and '
                             'put in <outdir> passed in via the command line and this tool '
                             'will save the hierarchy to NDEx using --ndexserver, --ndexuser, and '
                             '--ndexpassword credentials. If set to convert, it is assumes hierarchy has been generated'
                             ' (named hierarchy.cx2) and it converts the hierarchy to HiDeF .nodes and .edges files')
    parser.add_argument('--hcx_dir',
                        help='Input directory for convert mode with hierarchy in hcx to be converted to HiDeF .nodes '
                             'and .edges files')
    parser.add_argument('--provenance',
                        help='Path to file containing provenance '
                             'information about input files in JSON format. '
                             'This is required if inputdir does not contain '
                             'ro-crate-metadata.json file.')
    parser.add_argument('--name',
                        help='Name of this run, needed for FAIRSCAPE. If '
                             'unset, name value from specified '
                             'by --coembedding_dir directory or provenance file will be used')
    parser.add_argument('--organization_name',
                        help='Name of organization running this tool, needed '
                             'for FAIRSCAPE. If unset, organization name specified '
                             'in --coembedding_dir directory or provenance file will be used')
    parser.add_argument('--project_name',
                        help='Name of project running this tool, needed for '
                             'FAIRSCAPE. If unset, project name specified '
                             'in --coembedding_dir directory or provenance file will be used')
    parser.add_argument('--k', default=CellmapsGenerateHierarchy.K_DEFAULT, type=float,
                        help='HiDeF stability parameter')
    parser.add_argument('--algorithm', default=CellmapsGenerateHierarchy.ALGORITHM,
                        help='HiDeF clustering algorithm parameter')
    parser.add_argument('--maxres', default=CellmapsGenerateHierarchy.MAXRES, type=float,
                        help='HiDeF max resolution parameter')
    parser.add_argument('--containment_threshold', default=HiDeFHierarchyRefiner.CONTAINMENT_THRESHOLD, type=float,
                        help='Containment index threshold for pruning hierarchy')
    parser.add_argument('--jaccard_threshold', default=HiDeFHierarchyRefiner.JACCARD_THRESHOLD, type=float,
                        help='Jaccard index threshold for merging similar clusters')
    parser.add_argument('--min_diff', default=HiDeFHierarchyRefiner.MIN_DIFF, type=float,
                        help='Minimum difference in number of proteins for every '
                             'parent-child pair')
    parser.add_argument('--min_system_size', default=HiDeFHierarchyRefiner.MIN_SYSTEM_SIZE, type=float,
                        help='Minimum number of proteins each system must have to be kept')
    parser.add_argument('--ppi_cutoffs', nargs='+', type=float,
                        default=CosineSimilarityPPIGenerator.PPI_CUTOFFS,
                        help='Cutoffs used to generate PPI input networks. For example, '
                             'a value of 0.1 means to generate PPI input network using the '
                             'top ten percent of coembedding entries. Each cutoff generates '
                             'another PPI network')
    parser.add_argument('--hierarchy_parent_cutoff',
                        default=CDAPSHiDeFHierarchyGenerator.HIERARCHY_PARENT_CUTOFF, type=float,
                        help='PPI network cutoff to be chosen as hierarchy parent network.')
    parser.add_argument('--bootstrap_edges', type=validate_percentage,
                        default=CDAPSHiDeFHierarchyGenerator.BOOTSTRAP_EDGES,
                        help='Percentage of edges that will be removed randomly for bootstrapping, up to 99.')
    parser.add_argument('--skip_layout', action='store_true',
                        help='If set, skips layout of hierarchy step')
    parser.add_argument('--ndexserver', default='ndexbio.org',
                        help='Server where hierarchy can be converted to HCX and saved')
    parser.add_argument('--ndexuser',
                        help='NDEx user account')
    parser.add_argument('--ndexpassword',
                        help='NDEx password. Enter "-" to input password interactively, '
                             'or provide a file containing the password. Leave blank to not use a password.')
    parser.add_argument('--visibility', action='store_true',
                        help='If set, makes Hierarchy and interactome network loaded onto '
                             'NDEx publicly visible')
    parser.add_argument('--keep_intermediate_files', action='store_true',
                        help='If set, ppi network cx files will be saved.')
    parser.add_argument('--gene_node_attributes', nargs="+",
                        help='Accepts ro-crates that are output of imagedownloader or ppidownloader, '
                             'or tsv files with gene node attributes')
    parser.add_argument('--skip_logging', action='store_true',
                        help='If set, output.log, error.log '
                             'files will not be created')
    parser.add_argument('--logconf', default=None,
                        help='Path to python logging configuration file in '
                             'this format: https://docs.python.org/3/library/'
                             'logging.config.html#logging-config-fileformat '
                             'Setting this overrides -v parameter which uses '
                             ' default logger. (default None)')
    parser.add_argument('--verbose', '-v', action='count', default=1,
                        help='Increases verbosity of logger to standard '
                             'error for log messages in this module. Messages are '
                             'output at these python logging levels '
                             '-v = WARNING, -vv = INFO, '
                             '-vvv = DEBUG, -vvvv = NOTSET (default ERROR '
                             'logging)')
    parser.add_argument('--version', action='version',
                        version=('%(prog)s ' +
                                 cellmaps_generate_hierarchy.__version__))

    return parser.parse_args(args)


def validate_percentage(value):
    f_value = float(value)
    if f_value < 0 or f_value > 99:
        raise argparse.ArgumentTypeError(f"{value} is an invalid percentage value for --bootstrap_edges parameter. "
                                         f"Must be between 0 and 99")
    return f_value


def main(args):
    """
    Main entry point for program

    :param args: arguments passed to command line usually :py:func:`sys.argv[1:]`
    :type args: list

    :return: return value of :py:meth:`cellmaps_generate_hierarchy.runner.CellmapsGenerateHierarchy.run`
             or ``2`` if an exception is raised
    :rtype: int
    """
    desc = """
    Version {version}

    Takes a list of coembedding file {coembedding_file} files from {coembedding_dirs} directories
    (corresponding to multiple folds of the same data) that is in TSV format and generates several interaction networks
    that are fed via -g flag to HiDeF to create a hierarchy.

    Format of {coembedding_file} where 1st line is header:

    ''\t1\t2\t3\t4\t5...1024
    GENESYMBOL\tEMBEDDING1\tEMBEDDING2...

    Example:

            1       2       3       4       5
    AAAS    -0.35026753     -0.1307554      -0.046265163    0.3758623       0.22126552

    """.format(version=cellmaps_generate_hierarchy.__version__,
               coembedding_file=constants.CO_EMBEDDING_FILE,
               coembedding_dirs=', '.join(CO_EMBEDDINGDIRS))
    theargs = _parse_arguments(desc, args[1:])
    theargs.program = args[0]
    theargs.version = cellmaps_generate_hierarchy.__version__

    if theargs.provenance is not None:
        with open(theargs.provenance, 'r') as f:
            json_prov = json.load(f)
    else:
        json_prov = None

    try:
        logutils.setup_cmd_logging(theargs)
        if theargs.ndexpassword == '-':
            theargs.ndexpassword = getpass.getpass(prompt="Enter NDEx Password: ")
        if theargs.mode == 'ndexsave':
            ndex_uploader = NDExHierarchyUploader(theargs.ndexserver, theargs.ndexuser, theargs.ndexpassword,
                                                  theargs.visibility)
            _, _, _, hierarchyurl = ndex_uploader.upload_hierarchy_and_parent_network_from_files(theargs.outdir)
            print(f'Hierarchy uploaded. To view hierarchy on NDEx please paste this URL in your '
                  f'browser {hierarchyurl}. To view Hierarchy on new experimental Cytoscape on the Web, go to '
                  f'{ndex_uploader.get_cytoscape_url(hierarchyurl)}')
            return 0
        if theargs.mode == 'convert':
            hcx_dir = theargs.hcx_dir if theargs.hcx_dir is not None else theargs.outdir
            if not os.path.isdir(theargs.outdir):
                os.makedirs(theargs.outdir, mode=0o755)
            hidef_converter = HierarchyToHiDeFConverter(theargs.outdir, input_dir=hcx_dir)
            return hidef_converter.generate_hidef_files()

        if theargs.coembedding_dirs is None:
            raise CellmapsGenerateHierarchyError('In run mode, coembedding_dirs parameter is required.')

        provenance = ProvenanceUtil()
        ppigen = CosineSimilarityPPIGenerator(embeddingdirs=theargs.coembedding_dirs,
                                              cutoffs=theargs.ppi_cutoffs)

        refiner = HiDeFHierarchyRefiner(ci_thre=theargs.containment_threshold,
                                        ji_thre=theargs.jaccard_threshold,
                                        min_term_size=theargs.min_system_size,
                                        min_diff=theargs.min_diff,
                                        provenance_utils=provenance)

        converter = HCXFromCDAPSCXHierarchy()

        hiergen = CDAPSHiDeFHierarchyGenerator(author='cellmaps_generate_hierarchy',
                                               refiner=refiner,
                                               hcxconverter=converter,
                                               hierarchy_parent_cutoff=float(theargs.hierarchy_parent_cutoff),
                                               version=cellmaps_generate_hierarchy.__version__,
                                               provenance_utils=provenance,
                                               bootstrap_edges=theargs.bootstrap_edges)
        if theargs.skip_layout is True:
            layoutalgo = None
        else:
            layoutalgo = CytoscapeJSBreadthFirstLayout()

        # we dont want to log the password anywhere so toss it from the dict
        input_data_dict = theargs.__dict__.copy()
        if 'ndexpassword' in input_data_dict:
            input_data_dict['ndexpassword'] = 'PASSWORD REMOVED FOR SECURITY REASONS'

        return CellmapsGenerateHierarchy(outdir=theargs.outdir,
                                         inputdirs=theargs.coembedding_dirs,
                                         ppigen=ppigen,
                                         algorithm=theargs.algorithm,
                                         maxres=theargs.maxres,
                                         k=theargs.k,
                                         gene_node_attributes=theargs.gene_node_attributes,
                                         hiergen=hiergen,
                                         name=theargs.name,
                                         project_name=theargs.project_name,
                                         organization_name=theargs.organization_name,
                                         layoutalgo=layoutalgo,
                                         skip_logging=theargs.skip_logging,
                                         input_data_dict=input_data_dict,
                                         provenance_utils=provenance,
                                         ndexserver=theargs.ndexserver,
                                         ndexuser=theargs.ndexuser,
                                         ndexpassword=theargs.ndexpassword,
                                         visibility=theargs.visibility,
                                         keep_intermediate_files=theargs.keep_intermediate_files,
                                         provenance=json_prov
                                         ).run()
    except Exception as e:
        logger.exception('Caught exception: ' + str(e))
        return 2
    finally:
        logging.shutdown()


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main(sys.argv))



import os
import math
import pandas as pd
import numpy as np
import ndex2
from cellmaps_utils import music_utils
from cellmaps_utils import constants
from cellmaps_generate_hierarchy.exceptions import CellmapsGenerateHierarchyError


class PPINetworkGenerator(object):
    """
    Base class for objects that generate
    Protein to Protein interaction networks
    """

    def __init__(self):
        """
        Constructor
        """
        pass

    def get_next_network(self):
        """
        Gets next protein to protein interaction network

        :return: Network
        :rtype: :py:class:`ndex2.nice_cx_network.NiceCXNetwork`
        """
        raise NotImplementedError('subclasses need to implement')


class CosineSimilarityPPIGenerator(PPINetworkGenerator):
    """
    Takes Embedding file of format:

    .. code-block::

        ID # # # #

    Where ID is gene and #'s is embedding vector
    """
    PPI_CUTOFFS = [0.001, 0.002, 0.003, 0.004, 0.005, 0.006,
                   0.007, 0.008, 0.009, 0.01, 0.02, 0.03,
                   0.04, 0.05, 0.10]

    def __init__(self, embeddingdirs=[],
                 cutoffs=PPI_CUTOFFS):
        """
        Constructor
        """
        super().__init__()
        if embeddingdirs is None or len(embeddingdirs) < 1:
            raise CellmapsGenerateHierarchyError('embeddingdir is None')

        self._embeddingdirs = embeddingdirs
        self._cutoffs = cutoffs

    def _get_ppi_dataframe(self):
        """

        :return:
        """
        sim_mats = []
        index = []

        for fold, embeddingdir in enumerate(self._embeddingdirs):
            embeddingfile = os.path.join(embeddingdir, constants.CO_EMBEDDING_FILE)
            if os.path.exists(os.path.join(embeddingdir, constants.PPI_EMBEDDING_FILE)):
                embeddingfile = os.path.join(embeddingdir, constants.PPI_EMBEDDING_FILE)
            elif os.path.exists(os.path.join(embeddingdir, constants.IMAGE_EMBEDDING_FILE)):
                embeddingfile = os.path.join(embeddingdir, constants.IMAGE_EMBEDDING_FILE)

            z = pd.read_table(embeddingfile, sep='\t', index_col=0)

            # give the same ordering
            if len(index) == 0:
                index = z.index.values
            else:
                try:
                    z = z.loc[index]
                except KeyError:
                    index_set = set(index)
                    fold_set = set(z.index.values)
                    missing_in_current_fold = index_set.difference(fold_set)
                    extra_in_current_fold = fold_set.difference(index_set)
                    error_message = (
                        f"Discrepancy in protein sets across embedding folds detected. "
                        f"Proteins present in fold 1 but absent in fold {fold + 1}: {missing_in_current_fold}. "
                        f"Proteins present in fold {fold + 1} but absent in fold 1: {extra_in_current_fold}."
                    )
                    raise CellmapsGenerateHierarchyError(error_message)

            sim_mat = music_utils.cosine_similarity_scaled(z)
            keep = np.triu(np.ones(sim_mat.shape)).astype(bool)
            sim_mat = sim_mat.where(keep)
            sim_mats.append(sim_mat)

        # take mean across folds
        sim_mat = pd.DataFrame(np.array(sim_mats).mean(axis=0), index=index, columns=index)

        pairs = sim_mat.stack().reset_index().rename(columns={'level_0': constants.PPI_EDGELIST_GENEA_COL,
                                                              'level_1': constants.PPI_EDGELIST_GENEB_COL,
                                                              0: constants.WEIGHTED_PPI_EDGELIST_WEIGHT_COL})

        pairs = pairs[pairs[constants.PPI_EDGELIST_GENEA_COL] != pairs[constants.PPI_EDGELIST_GENEB_COL]]
        return pairs.sort_values(constants.WEIGHTED_PPI_EDGELIST_WEIGHT_COL, ascending=False)

    def get_next_network(self):
        """
        Gets all the edges

        :param cutoff: Fraction of top edges to keep
                       0.01 means 1% 0.5 means 50%
        :type cutoff: float
        :return: Network
        :rtype: :py:class:`ndex2.nice_cx_network.NiceCXNetwork`
        """
        df = self._get_ppi_dataframe()
        for cutoff in self._cutoffs:
            df_cutoff = df.iloc[0:math.ceil(cutoff * len(df))]
            net = ndex2.create_nice_cx_from_pandas(df_cutoff,
                                                   source_field=constants.PPI_EDGELIST_GENEA_COL,
                                                   target_field=constants.PPI_EDGELIST_GENEB_COL,
                                                   edge_attr=[constants.WEIGHTED_PPI_EDGELIST_WEIGHT_COL])
            net.set_name('parent interactome with ' + str(cutoff) + ' cutoff')
            net.set_network_attribute(name='description',
                                      values='Protein to Protein Interaction\n'
                                             'network generated by cellmaps_generate_hierarchy\n'
                                             'tool where top ' +
                                             str(round(cutoff * 100.0)) +
                                             '% of interactions sorted by weight\n')
            net.set_network_attribute(name='cutoff', values=str(cutoff))
            yield net



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `CosineSimilarityPPIGenerator`."""

import os

import shutil
import tempfile
import unittest
from cellmaps_utils import constants
from cellmaps_generate_hierarchy.runner import CellmapsGenerateHierarchy
from cellmaps_generate_hierarchy.ppi import CosineSimilarityPPIGenerator


class TestCosineSimilarityPPIGenerator(unittest.TestCase):
    """Tests for `CosineSimilarityPPIGenerator`."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def get_fake_five_line_coembedding_file(self):
        return os.path.join(os.path.dirname(__file__), 'data',
                            'fake_4_node_coembedding.tsv')

    def test_fake_five_line_coembedding(self):
        temp_dir = tempfile.mkdtemp()
        try:
            shutil.copy(self.get_fake_five_line_coembedding_file(),
                        os.path.join(temp_dir, constants.CO_EMBEDDING_FILE))
            gen = CosineSimilarityPPIGenerator(embeddingdirs=[temp_dir],
                                               cutoffs=[1.0])

            itr = gen.get_next_network()
            net = next(itr, None)
            self.assertEqual('parent interactome with 1.0 cutoff', net.get_name())
            self.assertEqual(6, len(net.get_edges()))
            self.assertEqual(4, len(net.get_nodes()))
            self.assertIsNone(next(itr, None))
        finally:
            shutil.rmtree(temp_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `CytoscapeJSBreadthFirstLayout`."""

import os
import shutil
import tempfile
import unittest
from unittest.mock import MagicMock
import requests
import time
import requests_mock
import ndex2
from cellmaps_generate_hierarchy.layout import CytoscapeJSBreadthFirstLayout
from cellmaps_generate_hierarchy.exceptions import CellmapsGenerateHierarchyError

class TestMatureHierarchy(unittest.TestCase):
    """Tests for `CDAPSHierarchyGenerator`."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_add_layout_500_status_code(self):
        runner = CytoscapeJSBreadthFirstLayout(rest_endpoint='http://xx',
                                               retry_sleep_time=0)
        net = ndex2.nice_cx_network.NiceCXNetwork()
        with requests_mock.Mocker() as m:
            m.post('http://xx', status_code=500)
            try:
                runner.add_layout(network=net)
                self.fail('Expected exception')
            except CellmapsGenerateHierarchyError as ce:
                self.assertTrue('Error running layout' in str(ce))

    def test_add_layout_task_failed(self):
        runner = CytoscapeJSBreadthFirstLayout(rest_endpoint='http://xx',
                                               retry_sleep_time=0)
        net = ndex2.nice_cx_network.NiceCXNetwork()
        with requests_mock.Mocker() as m:
            m.post('http://xx', status_code=202, json={'id': '12345'})
            m.get('http://xx/12345/status', status_code=200,
                  json={'progress': 100, 'status': 'error'})
            try:
                runner.add_layout(network=net)
                self.fail('Expected exception')
            except CellmapsGenerateHierarchyError as ce:
                self.assertTrue('Task failed' in str(ce))

    def test_add_layout_task_timeout(self):
        runner = CytoscapeJSBreadthFirstLayout(rest_endpoint='http://xx',
                                               retry_sleep_time=1)
        net = ndex2.nice_cx_network.NiceCXNetwork()
        with requests_mock.Mocker() as m:
            m.post('http://xx', status_code=202, json={'id': '12345'})
            m.get('http://xx/12345/status', status_code=500,
                  json={'progress': 100, 'status': 'error'})
            try:
                runner.add_layout(network=net, timeout=0)
                self.fail('Expected exception')
            except CellmapsGenerateHierarchyError as ce:
                self.assertTrue('Layout task exceeded' in str(ce))

    def test_add_layout_error_getting_final_result(self):
        runner = CytoscapeJSBreadthFirstLayout(rest_endpoint='http://xx',
                                               retry_sleep_time=0)
        net = ndex2.nice_cx_network.NiceCXNetwork()
        with requests_mock.Mocker() as m:
            m.post('http://xx', status_code=202, json={'id': '12345'})
            m.get('http://xx/12345/status', status_code=200,
                  json={'progress': 100, 'status': 'complete'})
            m.get('http://xx/12345', status_code=500, json={})
            try:
                runner.add_layout(network=net, timeout=0)
                self.fail('Expected exception')
            except CellmapsGenerateHierarchyError as ce:
                self.assertTrue('Error getting layout result' in str(ce))

    def test_add_layout_success(self):
        runner = CytoscapeJSBreadthFirstLayout(rest_endpoint='http://xx',
                                               retry_sleep_time=1)
        net = ndex2.nice_cx_network.NiceCXNetwork()
        with requests_mock.Mocker() as m:
            m.post('http://xx', status_code=202, json={'id': '12345'})
            m.get('http://xx/12345/status', status_code=200,
                  json={'progress': 100, 'status': 'complete'})
            m.get('http://xx/12345', status_code=200, json={'result': {'hi': 'there'}})
            runner.add_layout(network=net, timeout=0)
            self.assertEqual([{'hi': 'there'}], net.get_opaque_aspect('cartesianLayout'))

    def test_get_id_from_response_noid(self):
        runner = CytoscapeJSBreadthFirstLayout()
        try:
            mockres = MagicMock()
            mockres.json = MagicMock(return_val={})
            runner._get_id_from_response(mockres)
            self.fail('Expected exception')
        except CellmapsGenerateHierarchyError as ce:
            self.assertTrue('Error getting id from ' in str(ce))

    def test_is_task_complete_timeout_exceeded(self):
        runner = CytoscapeJSBreadthFirstLayout(retry_sleep_time=0)
        try:
            mockres = MagicMock()
            mockres.json = MagicMock(return_val={'progress': 50})
            start_time = int(time.time())-2
            runner._is_task_complete(mockres,
                                     start_time=start_time,
                                     timeout=1)
            self.fail('Expected exception')
        except CellmapsGenerateHierarchyError as ce:
            self.assertTrue('Layout task exceeded' in str(ce))

    def test_is_task_complete_notdone(self):
        runner = CytoscapeJSBreadthFirstLayout(retry_sleep_time=0)
        mockres = MagicMock()
        mockres.json = MagicMock(return_val={'progress': 50})
        start_time = int(time.time())
        runner._is_task_complete(mockres,
                                 start_time=start_time,
                                 timeout=500)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `MatureHierarchy`."""

import os
import shutil
import tempfile
import unittest
import pandas as pd
from cellmaps_generate_hierarchy.maturehierarchy import HiDeFHierarchyRefiner


class TestMatureHierarchy(unittest.TestCase):
    """Tests for `CDAPSHierarchyGenerator`."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_jaccard(self):
        a = {1}
        b = {1}
        self.assertEqual(1.0, HiDeFHierarchyRefiner._jaccard(a, b))
        self.assertEqual(1.0, HiDeFHierarchyRefiner._jaccard(b, a))

        a = {1}
        b = {1, 2}
        self.assertEqual(0.5, HiDeFHierarchyRefiner._jaccard(list(a), list(b)))
        self.assertEqual(0.5, HiDeFHierarchyRefiner._jaccard(b, a))

    def test_get_node_table_from_hidef(self):
        temp_dir = tempfile.mkdtemp()
        try:
            nodes_file = os.path.join(temp_dir, 'foo.nodes')
            with open(nodes_file, 'w') as f:
                f.write('Cluster1-406\t4\t4669 649 700 809\t10\n')
                f.write('Cluster1-407\t4\t123 3146 3251 3991\t10\n')
                f.write('Cluster1-408\t4\t1528 2310 2666 4273\t10\n')
            node_table = HiDeFHierarchyRefiner._get_node_table_from_hidef(nodes_file)
            self.assertEqual(3, len(node_table))
            self.assertEqual({'Cluster1-406',
                              'Cluster1-407',
                              'Cluster1-408'}, set(node_table[HiDeFHierarchyRefiner.TERMS_COL].values))
            self.assertEqual({4}, set(node_table[HiDeFHierarchyRefiner.TSIZE_COL].values))
        finally:
            shutil.rmtree(temp_dir)

    def test_edge_table_from_hidef(self):
        temp_dir = tempfile.mkdtemp()
        try:
            edges_file = os.path.join(temp_dir, 'foo.edges')
            with open(edges_file, 'w') as f:
                f.write('Cluster1-43\tCluster2-4\tdefault\n')
                f.write('Cluster1-51\tCluster2-16\tdefault\n')
                f.write('Cluster1-66\tCluster2-8\tdefault\n')

            edge_table = HiDeFHierarchyRefiner._get_edge_table_from_hidef(edges_file)
            self.assertEqual(3, len(edge_table))
            self.assertEqual({'Cluster1-43',
                               'Cluster1-51',
                               'Cluster1-66'}, set(edge_table[HiDeFHierarchyRefiner.PARENT_COL].values))

        finally:
            shutil.rmtree(temp_dir)

    def test_get_node_table_filtered_by_term_size(self):
        node_table = pd.DataFrame.from_dict({HiDeFHierarchyRefiner.TERMS_COL: ['one', 'two'],
                                             HiDeFHierarchyRefiner.TSIZE_COL: [3, 5]})

        res = HiDeFHierarchyRefiner._get_node_table_filtered_by_term_size(node_table)
        self.assertEqual(1, len(res))
        self.assertEqual({'two'}, set(res[HiDeFHierarchyRefiner.TERMS_COL].values))

        res = HiDeFHierarchyRefiner._get_node_table_filtered_by_term_size(node_table,
                                                                          min_term_size=3)
        self.assertEqual(2, len(res))
        self.assertEqual({'one', 'two'}, set(res[HiDeFHierarchyRefiner.TERMS_COL].values))

    def test_get_edge_table_filtered_by_node_set(self):
        edge_table = pd.DataFrame.from_dict({HiDeFHierarchyRefiner.PARENT_COL: ['Cluster1-43', 'Cluster1-51'],
                                             HiDeFHierarchyRefiner.CHILD_COL: ['Cluster2-4', 'Cluster2-16']})

        res = HiDeFHierarchyRefiner._get_edge_table_filtered_by_node_set(edge_table,
                                                                         node_set={'Cluster1-43', 'Cluster2-4'})
        self.assertEqual(1, len(res))

        res = HiDeFHierarchyRefiner._get_edge_table_filtered_by_node_set(edge_table,
                                                                         node_set={'Cluster1-43'})
        self.assertEqual(0, len(res))

    def test_get_leaves_from_edge_table(self):
        edge_table = pd.DataFrame.from_dict({HiDeFHierarchyRefiner.PARENT_COL: ['Cluster1-1', 'Cluster1-2'],
                                             HiDeFHierarchyRefiner.CHILD_COL: ['Cluster1-2', 'Cluster1-3']})
        res = HiDeFHierarchyRefiner._get_leaves_from_edge_table(edge_table)
        self.assertEqual({'Cluster1-3'}, res)

    def get_genes_from_node_table_for_term(self):
        node_table = pd.DataFrame.from_dict({HiDeFHierarchyRefiner.TERMS_COL: ['one', 'two'],
                                             HiDeFHierarchyRefiner.GENES_COL: ['genea geneb genec',
                                                         'gened genee genef']})

        res = HiDeFHierarchyRefiner._get_genes_from_node_table_for_term(node_table, 'one')
        self.assertEqual(['genea', 'geneb', 'genec'], res)

    def test_get_leaf_genes_to_add(self):
        node_table = pd.DataFrame.from_dict({HiDeFHierarchyRefiner.TERMS_COL: ['one', 'two'],
                                             HiDeFHierarchyRefiner.GENES_COL: ['genea geneb genec',
                                                         'gened genee genef']})
        refiner = HiDeFHierarchyRefiner()
        res = refiner._get_leaf_genes_to_add(node_table=node_table,
                                             leaves=['two'])
        self.assertEqual(3, len(res))
        self.assertTrue(['two', 'gened', HiDeFHierarchyRefiner.GENE_TYPE] in res)
        self.assertTrue(['two', 'genee', HiDeFHierarchyRefiner.GENE_TYPE] in res)
        self.assertTrue(['two', 'genef', HiDeFHierarchyRefiner.GENE_TYPE] in res)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `HcxHierarchy`."""
import unittest
from unittest.mock import patch, MagicMock
import ndex2
from ndex2.cx2 import CX2Network
import os

from cellmaps_generate_hierarchy.exceptions import CellmapsGenerateHierarchyError
from cellmaps_generate_hierarchy.hcx import HCXFromCDAPSCXHierarchy


class TestHcxHierarchy(unittest.TestCase):
    """Tests for `HcxHierarchy`."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    @staticmethod
    def _get_simple_nicecx_hierarchy():
        net = ndex2.nice_cx_network.NiceCXNetwork()
        root = net.create_node('root')
        child1 = net.create_node('child1')

        net.create_edge(edge_source=root,
                        edge_target=child1)

        child2 = net.create_node('child2')
        net.create_edge(edge_source=root,
                        edge_target=child2)

        subchild1 = net.create_node('subchild1')
        net.create_edge(edge_source=child1,
                        edge_target=subchild1)
        return net

    @staticmethod
    def _get_simple_cx2_hierarchy():
        net = ndex2.cx2.CX2Network()
        root = net.add_node(1, {'n': 'root'})
        child1 = net.add_node(2, {'n': 'child1'})

        net.add_edge(source=root, target=child1)

        child2 = net.add_node(3, {'n': 'child2'})
        net.add_edge(source=root, target=child2)

        subchild1 = net.add_node(4, {'n': 'subchild1'})
        net.add_edge(source=child1, target=subchild1)
        return net

    def test_get_root_nodes(self):
        net = self._get_simple_nicecx_hierarchy()
        myobj = HCXFromCDAPSCXHierarchy()
        self.assertEqual({0}, myobj._get_root_nodes(net))

    def test_add_isroot_node_attribute(self):
        net = self._get_simple_nicecx_hierarchy()
        myobj = HCXFromCDAPSCXHierarchy()
        myobj._add_isroot_node_attribute(net, root_nodes={0})
        self.assertEqual('true', net.get_node_attribute(0, 'HCX::isRoot')['v'])
        for x in [1, 2, 3]:
            self.assertEqual('false', net.get_node_attribute(x, 'HCX::isRoot')['v'])

    def test_get_mapping_of_node_names_to_ids(self):
        net = self._get_simple_nicecx_hierarchy()
        myobj = HCXFromCDAPSCXHierarchy()
        res = myobj._get_mapping_of_node_names_to_ids(net)
        self.assertEqual({'child1': 1, 'child2': 2,
                          'root': 0, 'subchild1': 3}, res)

    def test_add_hierarchy_network_attributes(self):
        net = self._get_simple_nicecx_hierarchy()
        myobj = HCXFromCDAPSCXHierarchy()
        myobj._add_hierarchy_network_attributes(net, interactome_name='foo_name')
        self.assertEqual('hierarchy_v0.1',
                         net.get_network_attribute('ndexSchema')['v'])
        self.assertEqual('2',
                         net.get_network_attribute('HCX::modelFileCount')['v'])
        self.assertEqual('foo_name',
                         net.get_network_attribute('HCX::interactionNetworkName')['v'])

    def test_add_members_node_attribute(self):
        net = self._get_simple_nicecx_hierarchy()
        name_map = {'A': 100, 'B': 200}
        net.set_node_attribute(0, 'CD_MemberList', values='A B C')

        myobj = HCXFromCDAPSCXHierarchy()
        myobj._add_members_node_attribute(net, interactome_name_map=name_map)

        mem_list = net.get_node_attribute(0, 'HCX::members')['v']
        self.assertEqual(2, len(mem_list))
        self.assertTrue('100' in mem_list)
        self.assertTrue('200' in mem_list)

        self.assertEqual(None, net.get_node_attribute(1, 'HCX::members'))

    def test_add_members_node_attribute_map_is_none(self):
        net = self._get_simple_nicecx_hierarchy()
        myobj = HCXFromCDAPSCXHierarchy()
        try:
            myobj._add_members_node_attribute(net, interactome_name_map=None)
            self.fail('Expected exception')
        except CellmapsGenerateHierarchyError as he:
            self.assertEqual('interactome name map is None', str(he))

    def test_get_visual_editor_properties_aspect_from_network(self):
        myobj = HCXFromCDAPSCXHierarchy()

        # try with empty network
        empty_net = CX2Network()
        res = myobj._get_visual_editor_properties_aspect_from_network(network=empty_net)
        self.assertIsNone(res)

        # try with network containing the aspect
        fake_net = CX2Network()
        visual_edit_aspect = {HCXFromCDAPSCXHierarchy.VISUAL_EDITOR_PROPERTIES_ASPECT: 1}
        fake_net.set_opaque_aspects([visual_edit_aspect])
        res = myobj._get_visual_editor_properties_aspect_from_network(network=fake_net)
        self.assertEqual(res, visual_edit_aspect)



import os
import sys
import csv
import logging
import subprocess
from datetime import date
import random

import ndex2
import cdapsutil
import cellmaps_generate_hierarchy
from cellmaps_utils import constants
from cellmaps_utils.provenance import ProvenanceUtil
from cellmaps_generate_hierarchy.exceptions import CellmapsGenerateHierarchyError

logger = logging.getLogger(__name__)


class HierarchyGenerator(object):
    """
    Base class for generating hierarchy
    that is output in CX format following
    CDAPS style
    """

    def __init__(self,
                 provenance_utils=ProvenanceUtil(),
                 author='cellmaps_generate_hierarchy',
                 version=cellmaps_generate_hierarchy.__version__):
        """
        Constructor
        """
        self._provenance_utils = provenance_utils
        self._author = author
        self._version = version
        self._generated_dataset_ids = []

    def get_generated_dataset_ids(self):
        """
        Gets IDs of datasets created by this object
        that have been registered with FAIRSCAPE
        :return:
        """
        return self._generated_dataset_ids

    def get_hierarchy(self, networks, algorithm='leiden', maxres=80, k=10):
        """
        Gets hierarchy


        :return: (hierarchy as :py:class:`list`,
                  parent ppi as :py:class:`list`)
        :rtype: tuple
        """
        raise NotImplementedError('Subclasses need to implement')


class CDAPSHiDeFHierarchyGenerator(HierarchyGenerator):
    """
    Generates hierarchy using HiDeF
    """

    CDAPS_JSON_FILE = 'cdaps.json'

    EDGELIST_TSV = '.id.edgelist.tsv'

    HIDEF_OUT_PREFIX = 'hidef_output'

    TRANSLATED_HIDEF_OUT_PREFIX = 'hidefnames_output'

    CDRES_KEY_NAME = 'communityDetectionResult'

    NODE_CX_KEY_NAME = 'nodeAttributesAsCX2'

    ATTR_DEC_NAME = 'attributeDeclarations'

    PERSISTENCE_COL_NAME = 'HiDeF_persistence'

    HIERARCHY_PARENT_CUTOFF = 0.1

    BOOTSTRAP_EDGES = 0

    def __init__(self, hidef_cmd='hidef_finder.py',
                 provenance_utils=ProvenanceUtil(),
                 refiner=None,
                 hcxconverter=None,
                 hierarchy_parent_cutoff=HIERARCHY_PARENT_CUTOFF,
                 author='cellmaps_generate_hierarchy',
                 version=cellmaps_generate_hierarchy.__version__,
                 bootstrap_edges=BOOTSTRAP_EDGES):
        """

        :param hidef_cmd: HiDeF command line binary
        :type hidef_cmd: str
        :param provenance_utils:
        :param author:
        :type author: str
        :param version:
        """
        super().__init__(provenance_utils=provenance_utils,
                         author=author,
                         version=version)
        self._refiner = refiner
        self._hcxconverter = hcxconverter
        self._hierarchy_parent_cutoff = hierarchy_parent_cutoff
        self._python = sys.executable
        if os.sep not in hidef_cmd:
            self._hidef_cmd = os.path.join(os.path.dirname(self._python), hidef_cmd)
        else:
            self._hidef_cmd = hidef_cmd
        self._bootstrap_edges = bootstrap_edges

    def _get_max_node_id(self, nodes_file):
        """
        Examines the 'nodes_file' passed in and finds the value of
        highest node id.

        It is assumed the 'nodes_file' a tab delimited
        file of format:

        <CLUSTER NAME> <# NODES> <SPACE DELIMITED NODE IDS> <SCORE>

        :param nodes_file:
        :type nodes_file: Path to to nodes file from hidef output
        :return: highest node id found
        :rtype: int
        """
        maxval = None
        with open(nodes_file, 'r') as csvfile:
            linereader = csv.reader(csvfile, delimiter='\t')
            for row in linereader:
                for node in row[2].split(' '):
                    if maxval is None:
                        maxval = int(node)
                        continue
                    curval = int(node)
                    if curval > maxval:
                        maxval = curval
        return maxval

    def write_members_for_row(self, out_stream, row, cur_node_id):
        """
        Given a row from nodes file from hidef output the members
        of the clusters by parsing the <SPACE DELIMITED NODE IDS>
        as mentioned in :py:func:`#get_max_node_id` description.

        The output is written to `out_stream` for each node id
        in format:

        <cur_node_id>,<node id>,c-m;

        :param out_stream:
        :type out_stream: file like object
        :param row: Should be a line from hidef nodes file parsed
                    by :py:func:`csv.reader`
        :type row: iterator
        :param cur_node_id: id of cluster that contains the nodes
        :type cur_node_id: int
        :return: None
        """
        for node in row[2].split(' '):
            out_stream.write(str(cur_node_id) + ',' +
                             node + ',c-m;')

    def update_cluster_node_map(self, cluster_node_map, cluster, max_node_id):
        """
        Updates 'cluster_node_map' which is in format of

        <cluster name> => <node id>

        by adding 'cluster' to 'cluster_node_map' if it does not
        exist

        :param cluster_node_map: map of cluster names to node ids
        :type cluster_node_map: dict
        :param cluster: name of cluster
        :type cluster: str
        :param max_node_id: current max node id
        :type max_node_id: int
        :return: (new 'max_node_id' if 'cluster' was added otherwise 'max_node_id',
                  id corresponding to 'cluster' found in 'cluster_node_map')
        :rtype: tuple
        """
        if cluster not in cluster_node_map:
            max_node_id += 1
            cluster_node_map[cluster] = max_node_id
            cur_node_id = max_node_id
        else:
            cur_node_id = cluster_node_map[cluster]
        return max_node_id, cur_node_id

    def update_persistence_map(self, persistence_node_map, node_id, persistence_val):
        """

        :param persistence_node_map:
        :param node_id:
        :param persistence_val:
        :return:
        """
        if node_id not in persistence_node_map:
            persistence_node_map[node_id] = persistence_val

    def write_communities(self, out_stream, edge_file, cluster_node_map):
        """
        Writes out links between clusters in COMMUNITYDETECTRESULT format
        as noted in :py:func:`#convert_hidef_output_to_cdaps`

        using hidef edge file set in 'edge_file' that is expected to
        be in this tab delimited format:

        <SOURCE CLUSTER> <TARGET CLUSTER> <default>

        This function converts the <SOURCE CLUSTER> <TARGET CLUSTER>
        to new node ids (leveraging 'cluster_node_map')

        and writes the following output:

        <SOURCE CLUSTER NODE ID>,<TARGET CLUSTER NODE ID>,c-c;

        to the 'out_stream'

        :param out_stream: output stream
        :type out_stream: file like object
        :param edge_file: path to hidef edges file
        :type edge_file: str
        :return: None
        """
        with open(edge_file, 'r') as csvfile:
            linereader = csv.reader(csvfile, delimiter='\t')
            for row in linereader:
                out_stream.write(str(cluster_node_map[row[0]]) + ',' +
                                 str(cluster_node_map[row[1]]) + ',c-c;')
        out_stream.write('",')

    def write_persistence_node_attribute(self, out_stream, persistence_map):
        """

        :param out_stream:
        :param persistence_map:
        :return:
        """
        out_stream.write('"' + CDAPSHiDeFHierarchyGenerator.NODE_CX_KEY_NAME + '": {')
        out_stream.write('"' + CDAPSHiDeFHierarchyGenerator.ATTR_DEC_NAME + '": [{')
        out_stream.write('"nodes": { "' + CDAPSHiDeFHierarchyGenerator.PERSISTENCE_COL_NAME +
                         '": { "d": "integer", "a": "p1", "v": 0}}}],')
        out_stream.write('"nodes": [')
        is_first = True
        for key in persistence_map:
            if is_first is False:
                out_stream.write(',')
            else:
                is_first = False
            out_stream.write('{"id": ' + str(key) + ',')
            out_stream.write('"v": { "p1": ' + str(persistence_map[key]) + '}}')

        out_stream.write(']}}')

    def convert_hidef_output_to_cdaps(self, out_stream, outdir):
        """
        Looks for x.nodes and x.edges in `outdir` directory
        to generate output in COMMUNITYDETECTRESULT format:
        https://github.com/idekerlab/communitydetection-rest-server/wiki/COMMUNITYDETECTRESULT-format

        This method leverages

        :py:func:`#write_members_for_row`

        and

        :py:func:`#write_communities`

        to write output

        :param out_stream: output stream to write results
        :type out_stream: file like object
        :param outdir:
        :type outdir: str
        :return: None
        """
        nodefile = os.path.join(outdir,
                                CDAPSHiDeFHierarchyGenerator.HIDEF_OUT_PREFIX +
                                '.pruned.nodes')
        max_node_id = self._get_max_node_id(nodefile)
        cluster_node_map = {}
        persistence_map = {}
        out_stream.write('{"communityDetectionResult": "')
        with open(nodefile, 'r') as csvfile:
            linereader = csv.reader(csvfile, delimiter='\t')
            for row in linereader:
                max_node_id, cur_node_id = self.update_cluster_node_map(cluster_node_map,
                                                                        row[0],
                                                                        max_node_id)
                self.update_persistence_map(persistence_map, cur_node_id, row[-1])
                self.write_members_for_row(out_stream, row,
                                           cur_node_id)
        edge_file = os.path.join(outdir, CDAPSHiDeFHierarchyGenerator.HIDEF_OUT_PREFIX + '.pruned.edges')
        self.write_communities(out_stream, edge_file, cluster_node_map)
        self.write_persistence_node_attribute(out_stream, persistence_map)
        out_stream.write('\n')
        return None

    def _run_cmd(self, cmd, cwd=None, timeout=36000):
        """
        Runs command as a command line process

        :param cmd_to_run: command to run as list
        :type cmd_to_run: list
        :return: (return code, standard out, standard error)
        :rtype: tuple
        """
        logger.debug('Running command under ' + str(cwd) +
                     ' path: ' + str(cmd))
        p = subprocess.Popen(cmd, cwd=cwd,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        try:
            out, err = p.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            logger.warning('Timeout reached. Killing process')
            p.kill()
            out, err = p.communicate()
            raise CellmapsGenerateHierarchyError('Process timed out. exit code: ' +
                                                 str(p.returncode) +
                                                 ' stdout: ' + str(out) +
                                                 ' stderr: ' + str(err))

        return p.returncode, out, err

    def _get_largest_network(self, networks):
        """
        Finds largest network by file size

        :param networks: list of :py:class:`~ndex2.nice_cx_network.NiceCXNetwork` objects
        :type networks: list
        :return: Largest network
        :rtype: :py:class:`~ndex2.nice_cx_network.NiceCXNetwork`
        """
        largest_network = None
        max_file_size = 0
        for n in networks:
            file_size = os.path.getsize(n + constants.CX_SUFFIX)
            if file_size >= max_file_size:
                largest_network = n
                max_file_size = file_size
        return largest_network

    def _get_parent_net_with_specified_cutoff(self, network, path, closest_net, closest_path,
                                              min_difference=float('inf')):
        """
        Determines the network with cutoff closest to the hierarchy_parent_cutoff value.

        :param network: The network to be evaluated.
        :type network: Network
        :param path: The file path or identifier for the network.
        :type path: str
        :param closest_net: Currently identified the closest network.
        :type closest_net: Network
        :param closest_path: File path or identifier for the current closest network.
        :type closest_path: str
        :param min_difference: Minimum difference between the cutoff values, defaults to float('inf').
        :type min_difference: float
        :return: A tuple of the closest network, its path, and the minimum difference in cutoff values.
        :rtype: tuple
        """
        cutoff_attr = network.get_network_attribute('cutoff')
        cutoff_value = 1 if cutoff_attr is None else float(cutoff_attr['v'])
        if cutoff_value == self._hierarchy_parent_cutoff:
            return network, path, 0
        else:
            difference = abs(self._hierarchy_parent_cutoff - cutoff_value)

            if difference < min_difference:
                min_difference = difference
                closest_net = network
                closest_path = path

        return closest_net, closest_path, min_difference

    def _get_name_to_id_dict(self, network):
        """

        :param network:
        :return:
        """
        name_to_id = {}
        for node_id, node_obj in network.get_nodes():
            name_to_id[node_obj['n']] = node_id
        return name_to_id

    def _get_id_to_name_dict(self, network):
        """

        :param network:
        :return:
        """
        id_to_name = {}
        for node_id, node_obj in network.get_nodes():
            id_to_name[node_id] = node_obj['n']
        return id_to_name

    def _create_edgelist_files_for_networks(self, networks):
        """
        Iterates through **networks** prefix paths and loads the
        CX files. Method then creates a PREFIX_PATH
        :py:const:`CDAPSHiDeFHierarchyGenerator.EDGELIST_TSV`
        file for each network and returns those paths as a list

        :param networks: Prefix paths of input PPI networks
        :type networks: list
        :return: (parent network path,
                  :py:class:`~ndex2.nice_cx_network.NiceCXNetwork`,
                  largest network path,
                  :py:class:`list`)
        :rtype: tuple
        """
        net_paths = []

        largest_network_path = self._get_largest_network(networks)
        largest_network = ndex2.create_nice_cx_from_file(largest_network_path + constants.CX_SUFFIX)
        logger.debug('Largest network name: ' + largest_network.get_name())
        largest_name_to_id = self._get_name_to_id_dict(largest_network)

        # Bootstrap edges
        all_edges = [(edge_obj['s'], edge_obj['t']) for _, edge_obj in largest_network.get_edges()]
        num_edges_to_remove = int(len(all_edges) * (self._bootstrap_edges / 100))
        all_removed_edges = random.sample(all_edges, num_edges_to_remove)

        parent_net = None
        parent_path = None
        min_difference = float('inf')
        for n in networks:
            if largest_network_path == n:
                net = largest_network
            else:
                logger.debug('Creating NiceCXNetwork object from: ' + n + constants.CX_SUFFIX)
                net = ndex2.create_nice_cx_from_file(n + constants.CX_SUFFIX)
            dest_path = n + CDAPSHiDeFHierarchyGenerator.EDGELIST_TSV
            net_paths.append(dest_path)
            logger.debug('Writing out id edgelist: ' + str(dest_path))
            id_to_name = self._get_id_to_name_dict(net)

            remaining_edges = list()
            removed_edges = list()
            for _, edge_obj in net.get_edges():
                edge = (edge_obj['s'], edge_obj['t'])
                if len(removed_edges) >= int(len(net.get_edges()) * (self._bootstrap_edges / 100)):
                    remaining_edges.append(edge)
                elif edge not in all_removed_edges:
                    remaining_edges.append(edge)
                else:
                    removed_edges.append(edge)

            with open(dest_path, 'w') as f:
                for s, t in remaining_edges:
                    f.write(str(largest_name_to_id[id_to_name[s]]) + '\t' +
                            str(largest_name_to_id[id_to_name[t]]) + '\n')

            if len(removed_edges) > 0:
                removed_edges_path = n + '_removed_edges.tsv'
                with open(removed_edges_path, 'w') as f:
                    for s, t in removed_edges:
                        f.write(str(largest_name_to_id[id_to_name[s]]) + '\t' +
                                str(largest_name_to_id[id_to_name[t]]) + '\n')

            if len(remaining_edges) == 0:
                raise CellmapsGenerateHierarchyError(f"PPI network {n} has no edges. Cannot create hierarchy.")

            # register edgelist file with fairscape
            data_dict = {'name': os.path.basename(dest_path) + ' PPI id edgelist file',
                         'description': 'PPI id edgelist file',
                         'data-format': 'tsv',
                         'author': str(self._author),
                         'version': str(self._version),
                         'date-published': date.today().strftime(
                             self._provenance_utils.get_default_date_format_str())}
            dataset_id = self._provenance_utils.register_dataset(os.path.dirname(dest_path),
                                                                 source_file=dest_path,
                                                                 data_dict=data_dict)
            self._generated_dataset_ids.append(dataset_id)
            if min_difference != 0:
                parent_net, parent_path, min_difference = self._get_parent_net_with_specified_cutoff(
                    net, n, parent_net, parent_path, min_difference)

        logger.debug('Parent network name: ' + parent_net.get_name())
        return parent_path + constants.CX_SUFFIX, parent_net, largest_network, net_paths

    def _register_hidef_output_files(self, outdir):
        """
        Register <HIDEF_PREFIX>.nodes and <HIDEF_PREFIX>.edges
        and <HIDEF_PREFIX>.weaver files with FAIRSCAPE

        """

        for hidef_file in [('nodes', 'tsv'),
                           ('edges', 'tsv'),
                           ('weaver', 'npy')]:
            outfile = os.path.join(outdir,
                                   CDAPSHiDeFHierarchyGenerator.HIDEF_OUT_PREFIX +
                                   '.' + hidef_file[0])
            data_dict = {'name': os.path.basename(outfile) + ' HiDeF output ' + hidef_file[0] + ' file',
                         'description': ' HiDeF output ' + hidef_file[0] + ' file',
                         'data-format': hidef_file[1],
                         'author': str(self._author),
                         'version': str(self._version),
                         'date-published': date.today().strftime(self._provenance_utils.get_default_date_format_str())}
            dataset_id = self._provenance_utils.register_dataset(os.path.dirname(outfile),
                                                                 source_file=outfile,
                                                                 data_dict=data_dict)
            self._generated_dataset_ids.append(dataset_id)

    def _annotate_hierarchy(self, network=None, path=None):
        """
        Adds HCX attributes to network as well as sets

        ``prov:wasGeneratedBy`` to the name and version of this tool

        ``prov:wasDerivedFrom`` to FAIRSCAPE dataset id of this rocrate

        :param network: Hierarchy
        :type network: :py:class:`~ndex2.nice_cx_network.NiceCXNetwork`
        :param path: Path to parent PPI network in CX or CX2 format
        :type path: str
        """
        network.set_network_attribute(name='prov:wasGeneratedBy',
                                      values=self._author + ' ' + self._version)
        rocrate_id = self._provenance_utils.get_id_of_rocrate(os.path.dirname(path))

        description = 'Cell Map Hierarchy'
        if rocrate_id is not None:
            network.set_network_attribute(name='prov:wasDerivedFrom',
                                          values='RO-crate: ' + str(rocrate_id))

            prov_utils = self._provenance_utils.get_rocrate_provenance_attributes(os.path.dirname(path))
            if self._bootstrap_edges > 0:
                description = (description + ' derived from edgeslists with ' + str(self._bootstrap_edges) +
                               '% of edges randomly removed')
            network.set_network_attribute(name='description',
                                          values=description + '|' + str(prov_utils.get_description()))
            if prov_utils.get_keywords() is None:
                keyword_subset = []
            else:
                keyword_subset = prov_utils.get_keywords()[:6]
            network.set_name(prov_utils.get_name() + ' - ' + ' '.join(keyword_subset) + ' hierarchy')
        else:
            network.set_network_attribute(name='description', values=description)
            network.set_name(description)

    def _annotate_hierarchy_nodes(self, network):
        """
        Annotates each node in the hierarchy with its community name and a label flag.

        :param network: The hierarchy containing nodes to be annotated.
        :type network: :py:class:`~ndex2.nice_cx_network.NiceCXNetwork`
        """
        for node_id, node_obj in network.get_nodes():
            name = node_obj['n']
            network.set_node_attribute(node_id, 'CD_CommunityName',
                                       values=name, type='string',
                                       overwrite=True)
            network.set_node_attribute(node_id, 'CD_Labeled',
                                       values='true', type='boolean',
                                       overwrite=True)

    def _run_hidef(self, edgelist_files, outputprefix, algorithm, maxres, k):
        cmd = [self._python, self._hidef_cmd, '--g']
        cmd.extend(edgelist_files)
        cmd.extend(['--o', outputprefix,
                    '--alg', algorithm, '--maxres', str(maxres), '--k', str(k),
                    '--skipgml'])

        exit_code, out, err = self._run_cmd(cmd)

        if exit_code != 0:
            logger.error('Cmd failed with exit code: ' + str(exit_code) +
                         ' : ' + str(out) + ' : ' + str(err))
            raise CellmapsGenerateHierarchyError('Cmd failed with exit code: ' + str(exit_code) +
                                                 ' : ' + str(out) + ' : ' + str(err))

    def get_hierarchy_from_edgelists(self, outdir, edgelist_files, parent_net, algorithm='leiden', maxres=80, k=10):
        """
        Generates a hierarchy from edgelist files using HiDeF.

        This method runs the HiDeF algorithm on the provided edgelist files to generate a hierarchical community structure.
        It optionally refines the hierarchy, converts the HiDeF output to CDAPS format, and then uses `cdapsutil` to run
        community detection on the parent network.

        :param outdir: The output directory where HiDeF results and intermediate files will be stored.
        :type outdir: str
        :param edgelist_files: A list of paths to edgelist files to be used as input to HiDeF.
        :type edgelist_files: list
        :param parent_net: The parent network on which community detection is performed.
        :type parent_net: :py:class:`~ndex2.nice_cx_network.NiceCXNetwork` or :py:class:`~ndex2.cx2.CX2Network`
        :param algorithm: The algorithm to use for community detection (default is 'leiden').
        :type algorithm: str
        :param maxres: The maximum resolution parameter for HiDeF (default is 80).
        :type maxres: int
        :param k: The k parameter for HiDeF (default is 10).
        :type k: int
        :return: A tuple containing the resulting hierarchy and the path to the CDAPS output JSON file, or (None, None) if an error occurs.
        :rtype: tuple (hierarchy, str) or (None, None)
        :raises FileNotFoundError: If no output is generated from HiDeF.
        """
        outputprefix = os.path.join(outdir, CDAPSHiDeFHierarchyGenerator.HIDEF_OUT_PREFIX)
        self._run_hidef(edgelist_files, outputprefix, algorithm, maxres, k)

        try:
            if self._refiner is not None:
                self._refiner.refine_hierarchy(outprefix=outputprefix)

            cdaps_out_file = os.path.join(outdir,
                                          CDAPSHiDeFHierarchyGenerator.CDAPS_JSON_FILE)
            with open(cdaps_out_file, 'w') as out_stream:
                self.convert_hidef_output_to_cdaps(out_stream, outdir)

            cd = cdapsutil.CommunityDetection(runner=cdapsutil.ExternalResultsRunner())
            hier = cd.run_community_detection(parent_net, algorithm=cdaps_out_file)
            return hier, cdaps_out_file
        except FileNotFoundError as fe:
            logger.error('No output from hidef: ' + str(fe) + '\n')
        return None, None

    def get_hierarchy(self, networks, algorithm='leiden', maxres=80, k=10):
        """
        Runs HiDeF to generate hierarchy and registers resulting output
        files with FAIRSCAPE. To do this the method generates edgelist
        files from the CX files corresponding to the **networks** using
        the internal node ids for edge source and target names. These
        files are written to the same directory as the **networks**
        with HiDeF
        is then given all these networks via ``--g`` flag.



        .. warning::

            Due to FAIRSCAPE registration this method is NOT threadsafe and
            cannot be called in parallel or with any other call that is
            updating FAIRSCAPE registration on the current RO-CRATE

        :param networks: Paths (without suffix ie .cx) to PPI networks to be
                         used as input to HiDeF
        :type networks: list
        :param algorithm: The algorithm to use for community detection (default is 'leiden').
        :type algorithm: str
        :param maxres: The maximum resolution parameter for HiDeF (default is 80).
        :type maxres: int
        :param k: The k parameter for HiDeF (default is 10).
        :type k: int
        :raises CellmapsGenerateHierarchyError: If there was an error
        :return: Resulting hierarchy or ``None`` if no hierarchy from HiDeF
        :return: (hierarchy as list,
                  parent ppi as list,
                  hierarchyurl, parenturl)
                  or None, None if not created
        :rtype: tuple
        """
        if self._hcxconverter is None:
            raise CellmapsGenerateHierarchyError('HCX converter must be set')
        outdir = os.path.dirname(networks[0])

        (parent_net_path, parent_net, largest_net, edgelist_files) = self._create_edgelist_files_for_networks(networks)

        hier, cdaps_out_file = self.get_hierarchy_from_edgelists(outdir, edgelist_files, largest_net,
                                                                 algorithm, maxres, k)
        self._clean_tmp_edgelist_files(edgelist_files)
        self._annotate_hierarchy(network=hier, path=parent_net_path)
        self._annotate_hierarchy_nodes(network=hier)
        hierarchy_in_hcx = self._hcxconverter.get_converted_hierarchy(hierarchy=hier,
                                                                      parent_network=parent_net)

        # Register outputs from hierarchy generation
        self._register_hidef_output_files(outdir)

        # register cdaps json file with fairscape
        data_dict = {'name': os.path.basename(cdaps_out_file) + ' CDAPS output JSON file',
                     'description': 'CDAPS output JSON file',
                     'data-format': 'json',
                     'author': str(self._author),
                     'version': str(self._version),
                     'date-published': date.today().strftime(self._provenance_utils.get_default_date_format_str())}
        dataset_id = self._provenance_utils.register_dataset(os.path.dirname(cdaps_out_file),
                                                             source_file=cdaps_out_file,
                                                             data_dict=data_dict)
        self._generated_dataset_ids.append(dataset_id)

        return hierarchy_in_hcx

    def _clean_tmp_edgelist_files(self, edgelist_files):
        for file in edgelist_files:
            try:
                file_path = os.path.join(os.getcwd(), '_tmp.' + os.path.basename(file))
                if os.path.exists(file_path):
                    os.remove(file_path)
            except Exception as e:
                logger.warning(f"Tried to remove tmp file, but failed due to: {e}")




import logging
import requests
import time
import ndex2
from cellmaps_generate_hierarchy.exceptions import CellmapsGenerateHierarchyError
logger = logging.getLogger(__name__)


class HierarchyLayout(object):
    """
    Base class for layout algorithms
    """
    def __init__(self):
        """
        Constructor
        """
        pass

    def add_layout(self, network=None):
        """
        Adds layout to network passed in.
        Subclasses should implement

        :param network:
        :type network: :py:class:`~ndex2.nice_cx_network.NiceCXNetwork`
        :raises NotImplementedError: Always raised
        """
        raise NotImplementedError('Subclasses should implement')


class CytoscapeJSBreadthFirstLayout(HierarchyLayout):
    """
    Runs breadthfirst layout from http://cytolayouts.ucsd.edu/cd
    to get a layout
    """

    HEADERS = {'Content-Type': 'application/json',
               'Accept': 'application/json'}

    def __init__(self, layout_algorithm='breadthfirst',
                 rest_endpoint='http://cytolayouts.ucsd.edu/cd/communitydetection/v1',
                 retry_sleep_time=1,
                 request_timeout=120):
        """
        Constructor

        :param layout_algorithm: can be one of the following:
                                 circle|cose|grid|concentric|breadthfirst|dagre
        :type layout_algorithm: str
        :param rest_endpoint: URL for rest service
        :type request_timeout: str
        :param retry_sleep_time: time in seconds to wait before checking status
                                 with REST service on status of task
        :type retry_sleep_time: int or float
        :param request_timeout: timeout in seconds to pass to :py:mod:`requests`
                                library for web requests
        :type request_timeout: int or float
        """
        self._layout_algorithm = layout_algorithm
        self._rest_endpoint = rest_endpoint
        self._retry_sleep_time = retry_sleep_time
        self._request_timeout = request_timeout

    def add_layout(self, network=None,
                   timeout=1800):
        """
        Runs algorithm specified in constructor on **network**
        in place

        :param network: Hierarchy network
        :type network: :py:class:`~ndex2.nice_cx_network.NiceCXNetwork`
        :param timeout: time in seconds to wait for task to finish before
                        failing
        :type timeout: int or float
        """
        res = requests.post(self._rest_endpoint,
                            headers=CytoscapeJSBreadthFirstLayout.HEADERS,
                            json={'algorithm': 'cytojslayout',
                                  'customParameters': {'--layout': self._layout_algorithm},
                                  'data': network.to_cx()},
                            timeout=self._request_timeout)

        task_id = self._wait_for_task(res, timeout=timeout)

        res = requests.get(self._rest_endpoint + '/' + str(task_id),
                           headers=CytoscapeJSBreadthFirstLayout.HEADERS)

        if res.status_code != 200:
            raise CellmapsGenerateHierarchyError('Error getting layout results')

        network.set_opaque_aspect('cartesianLayout', res.json()['result'])

    def _get_task_status(self, task_id, start_time=None,
                         timeout=None):
        """
        Gets status of task

        :param task_id: id of task
        :type task_id: str
        :return: response
        :rtype: :py:class:`requests.Response`
        """
        logger.debug('Checking status of task: ' + str(task_id))
        res = requests.get(self._rest_endpoint + '/' +
                           str(task_id) + '/status',
                           headers=CytoscapeJSBreadthFirstLayout.HEADERS)
        if res.status_code != 200:
            logger.debug('Request came back with error, '
                         'sleeping ' + str(self._retry_sleep_time) +
                         ' second(s) and trying again : ' +
                         str(res.text))
            if (int(time.time()) - start_time) > timeout:
                raise CellmapsGenerateHierarchyError('Layout task exceeded timeout of ' +
                                                     str(timeout) + ' seconds')
            time.sleep(self._retry_sleep_time)
            return None
        return res

    def _get_id_from_response(self, res):
        """
        Extracts id of task from json response

        :param res: response from post to layout service
        :type res: :py:class`requests.Response`
        :raises CellmapsGenerateHierarchyError: If there is an error trying
                                                to get id from response
        :return: id of task
        :rtype: str
        """
        res_json = res.json()
        if 'id' not in res_json:
            raise CellmapsGenerateHierarchyError('Error getting id from response: ' +
                                                 str(res_json) + ' ' + str(res.text))
        return res_json['id']

    def _is_task_complete(self, res, start_time=None,
                          timeout=None):
        """
        Checks if task is complete by examining the json from the **res** and
        if ``progress`` is ``100`` and ``status`` is ``complete`` the task is done

        :param res:
        :param start_time: time task was started in seconds
        :type start_time: float
        :param timeout: Number of seconds task is allowed to run before
                        considering the task is a failure
        :type timeout: float
        :raises CellmapsGenerateHierarchyError: If task ``status`` is not ``complete`` or
                                                if **timeout** exceeded
        :return: ``True`` if task is completed otherwise ``False``
        :rtype: bool
        """
        status_dict = res.json()
        if status_dict['progress'] == 100:
            if status_dict['status'] != 'complete':
                raise CellmapsGenerateHierarchyError('Task failed: ' +
                                                     str(status_dict))
            return True
        if (int(time.time()) - start_time) > timeout:
            raise CellmapsGenerateHierarchyError('Layout task exceeded timeout of ' +
                                                 str(timeout) + ' seconds')
        time.sleep(self._retry_sleep_time)
        return False

    def _wait_for_task(self, res, timeout=1800):
        """
        Waits for task set in **res** response

        :param res:
        :type res: :py:class:`requests.Response`
        :param timeout: Time in seconds to wait for task to complete
        :type timeout: int
        :raises CellmapsGenerateHierarchyError: If there is an error
        :return: task id
        :rtype: str
        """
        if res.status_code != 202:
            raise CellmapsGenerateHierarchyError('Error running layout: ' +
                                                 str(res.status_code) + ' : ' +
                                                 str(res.text))

        task_id = self._get_id_from_response(res)

        start_time = int(time.time())
        complete = False
        while complete is False:
            res = self._get_task_status(task_id=task_id,
                                        start_time=start_time,
                                        timeout=timeout)
            if res is None:
                continue

            complete = self._is_task_complete(res, start_time=start_time,
                                              timeout=timeout)
        return task_id





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `cellmaps_generate_hierarchy` package."""

import os
import shutil
import tempfile
import json
import unittest
from unittest.mock import MagicMock, ANY

from cellmaps_utils.exceptions import CellMapsProvenanceError
from ndex2.cx2 import CX2Network

import cellmaps_generate_hierarchy
from cellmaps_utils import constants

from cellmaps_generate_hierarchy.exceptions import CellmapsGenerateHierarchyError
from cellmaps_generate_hierarchy.ndexupload import NDExHierarchyUploader
from cellmaps_generate_hierarchy.runner import CellmapsGenerateHierarchy


class TestCellmapsgeneratehierarchyrunner(unittest.TestCase):
    """Tests for `cellmaps_generate_hierarchy` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_constructor(self):
        """Tests constructor"""
        temp_dir = tempfile.mkdtemp()
        try:
            myobj = CellmapsGenerateHierarchy(outdir=os.path.join(temp_dir, 'out'))
            self.assertIsNotNone(myobj)
        finally:
            shutil.rmtree(temp_dir)

    def test_constructor_outdir_must_be_set(self):
        """Tests constructor outdir must be set"""
        try:
            CellmapsGenerateHierarchy()
            self.fail('Expected exception')
        except CellmapsGenerateHierarchyError as e:
            self.assertEqual('outdir is None', str(e))

    def test_run_without_logging(self):
        """ Tests run() without logging."""
        temp_dir = tempfile.mkdtemp()
        try:
            run_dir = os.path.join(temp_dir, 'run')
            myobj = CellmapsGenerateHierarchy(outdir=run_dir)
            try:
                myobj.run()
                self.fail('Expected CellmapsGenerateHierarchyError')
            except CellmapsGenerateHierarchyError as e:
                print(e)
                self.assertTrue('RO-Crate' in str(e))

            self.assertFalse(os.path.isfile(os.path.join(run_dir, 'output.log')))
            self.assertFalse(os.path.isfile(os.path.join(run_dir, 'error.log')))

        finally:
            shutil.rmtree(temp_dir)

    def test_run_with_logging(self):
        """ Tests run() with logging."""
        temp_dir = tempfile.mkdtemp()
        try:
            run_dir = os.path.join(temp_dir, 'run')
            myobj = CellmapsGenerateHierarchy(outdir=run_dir,
                                              skip_logging=False)
            try:
                myobj.run()
                self.fail('Expected CellmapsGenerateHierarchyError')
            except CellmapsGenerateHierarchyError as e:
                self.assertTrue('RO-Crate' in str(e))

            self.assertTrue(os.path.isfile(os.path.join(run_dir, 'output.log')))
            self.assertTrue(os.path.isfile(os.path.join(run_dir, 'error.log')))

        finally:
            shutil.rmtree(temp_dir)

    def test_create_rocrate(self):

        prov = MagicMock()
        prov.register_rocrate = MagicMock()
        gen = CellmapsGenerateHierarchy(outdir='/foo', inputdirs=[], name='crate name',
                                        organization_name='organization name',
                                        project_name='project name', provenance_utils=prov)
        gen._description = 'description'
        gen._create_rocrate()
        prov.register_rocrate.assert_called_with('/foo', name='crate name',
                                                 organization_name='organization name',
                                                 project_name='project name',
                                                 description='description',
                                                 keywords=None)

    def test_create_rocrate_raise_type_error(self):

        prov = MagicMock()
        prov.register_rocrate = MagicMock(side_effect=TypeError('uh'))
        try:
            gen = CellmapsGenerateHierarchy(outdir='/foo', provenance_utils=prov)
            gen._create_rocrate()
            self.fail('expected exception')
        except CellmapsGenerateHierarchyError as ce:
            self.assertTrue('Invalid provenance: uh' in str(ce))

    def test_create_rocrate_raise_key_error(self):

        prov = MagicMock()
        prov.register_rocrate = MagicMock(side_effect=KeyError('uh'))
        try:
            gen = CellmapsGenerateHierarchy(outdir='/foo', provenance_utils=prov)
            gen._create_rocrate()
            self.fail('expected exception')
        except CellmapsGenerateHierarchyError as ce:
            self.assertTrue('Key missing in provenance: ' in str(ce))

    def test_update_provenance_fields(self):
        prov = MagicMock()
        prov.get_merged_rocrate_provenance_attrs = MagicMock()
        gen = CellmapsGenerateHierarchy(outdir='/foo', inputdirs=['/crate1'],
                                        provenance_utils=prov)
        try:
            gen._update_provenance_fields()
            prov.get_merged_rocrate_provenance_attrs.assert_called_with(['/crate1'], override_name=None,
                                                                        override_project_name=None,
                                                                        override_organization_name=None,
                                                                        extra_keywords=['hierarchy', 'model'])
            self.fail('expected exception')
        except CellmapsGenerateHierarchyError as e:
            self.assertTrue('RO-Crate' in str(e))

    def test_register_software(self):

        prov = MagicMock()
        prov.register_software = MagicMock()
        gen = CellmapsGenerateHierarchy(outdir='/foo',
                                        provenance_utils=prov)
        gen._description = 'description'
        gen._keywords = ['hi']
        gen._register_software()
        software_description = gen._description + ' ' + cellmaps_generate_hierarchy.__description__
        prov.register_software.assert_called_once()
        self.assertEqual(('/foo',), prov.register_software.call_args.args)
        self.assertEqual('cellmaps_generate_hierarchy', prov.register_software.call_args.kwargs['name'])
        self.assertEqual(software_description, prov.register_software.call_args.kwargs['description'])
        self.assertEqual(cellmaps_generate_hierarchy.__author__, prov.register_software.call_args.kwargs['author'])
        self.assertEqual(cellmaps_generate_hierarchy.__version__, prov.register_software.call_args.kwargs['version'])
        self.assertEqual('py', prov.register_software.call_args.kwargs['file_format'])
        self.assertEqual(cellmaps_generate_hierarchy.__repo_url__, prov.register_software.call_args.kwargs['url'])
        self.assertEqual(3, len(prov.register_software.call_args.kwargs['keywords']))
        self.assertTrue('hi' in prov.register_software.call_args.kwargs['keywords'])
        self.assertTrue('tools' in prov.register_software.call_args.kwargs['keywords'])
        self.assertTrue('cellmaps_generate_hierarchy' in prov.register_software.call_args.kwargs['keywords'])

    def test_register_computation(self):

        prov = MagicMock()
        prov.register_computation = MagicMock()
        prov.get_login = MagicMock(return_value='smith')
        prov.get_id_of_rocrate = MagicMock(return_value='someid')
        gen = CellmapsGenerateHierarchy(outdir='/foo',
                                        provenance_utils=prov,
                                        input_data_dict={'foo': 'hi'})
        gen._inputdirs = '/xdir'
        gen._description = 'description'
        gen._keywords = ['hi']
        gen._softwareid = 'softid'
        gen._register_computation(['oneid'])

        description = gen._description + ' run of ' + cellmaps_generate_hierarchy.__name__
        prov.register_computation.assert_called_once()
        self.assertEqual(('/foo',), prov.register_computation.call_args.args)
        self.assertEqual('smith', prov.register_computation.call_args.kwargs['run_by'])
        self.assertEqual(description, prov.register_computation.call_args.kwargs['description'])
        self.assertEqual("{'foo': 'hi'}", prov.register_computation.call_args.kwargs['command'])
        self.assertEqual(['softid'], prov.register_computation.call_args.kwargs['used_software'])
        self.assertEqual(['someid'], prov.register_computation.call_args.kwargs['used_dataset'])
        self.assertEqual(['oneid'], prov.register_computation.call_args.kwargs['generated'])
        self.assertEqual(2, len(prov.register_computation.call_args.kwargs['keywords']))
        self.assertTrue('hi' in prov.register_computation.call_args.kwargs['keywords'])
        self.assertTrue('computation' in prov.register_computation.call_args.kwargs['keywords'])

    def test_register_computation_list_of_inputdirs(self):

        prov = MagicMock()
        prov.register_computation = MagicMock()
        prov.get_login = MagicMock(return_value='smith')
        prov.get_id_of_rocrate = MagicMock(side_effect=['1id', '2id'])
        gen = CellmapsGenerateHierarchy(outdir='/foo',
                                        provenance_utils=prov,
                                        input_data_dict={'foo': 'hi'})
        gen._inputdirs = ['/1', '/2']
        gen._description = 'description'
        gen._keywords = ['hi']
        gen._softwareid = 'softid'
        gen._register_computation(['oneid'])

        description = gen._description + ' run of ' + cellmaps_generate_hierarchy.__name__
        prov.register_computation.assert_called_once()
        self.assertEqual(('/foo',), prov.register_computation.call_args.args)
        self.assertEqual('smith', prov.register_computation.call_args.kwargs['run_by'])
        self.assertEqual(description, prov.register_computation.call_args.kwargs['description'])
        self.assertEqual("{'foo': 'hi'}", prov.register_computation.call_args.kwargs['command'])
        self.assertEqual(['softid'], prov.register_computation.call_args.kwargs['used_software'])
        self.assertEqual(['1id', '2id'], prov.register_computation.call_args.kwargs['used_dataset'])
        self.assertEqual(['oneid'], prov.register_computation.call_args.kwargs['generated'])
        self.assertEqual(2, len(prov.register_computation.call_args.kwargs['keywords']))
        self.assertTrue('hi' in prov.register_computation.call_args.kwargs['keywords'])
        self.assertTrue('computation' in prov.register_computation.call_args.kwargs['keywords'])

    def test_get_ppi_network_dest_file(self):
        ppi_net = MagicMock()
        ppi_net.get_network_attribute = MagicMock(return_value={'v': 1.5})
        gen = CellmapsGenerateHierarchy(outdir='/foo')
        self.assertEqual(os.path.join('/foo', constants.PPI_NETWORK_PREFIX + '_cutoff_1.5'),
                         gen.get_ppi_network_dest_file(ppi_network=ppi_net))

    def test_get_hierarchy_dest_file(self):
        gen = CellmapsGenerateHierarchy(outdir='/foo')
        self.assertEqual(os.path.join('/foo', constants.HIERARCHY_NETWORK_PREFIX),
                         gen.get_hierarchy_dest_file())

    def test_get_hierarchy_parent_network_dest_file(self):
        gen = CellmapsGenerateHierarchy(outdir='/foo')
        self.assertEqual(os.path.join('/foo', 'hierarchy_parent'),
                         gen.get_hierarchy_parent_network_dest_file())

    def test_remove_ppi_networks(self):
        temp_dir = tempfile.mkdtemp()
        try:
            afile = os.path.join(temp_dir, 'foo' + constants.CX_SUFFIX)
            open(afile, 'a').close()
            gen = CellmapsGenerateHierarchy(outdir=temp_dir)
            gen._remove_ppi_networks([os.path.join(temp_dir, 'foo'),
                                      os.path.join(temp_dir, 'doesnotexist')])
            self.assertFalse(os.path.isfile(afile))
        finally:
            shutil.rmtree(temp_dir)

    def test_write_ppi_network_as_cx(self):
        temp_dir = tempfile.mkdtemp()
        try:
            afile = os.path.join(temp_dir, 'hi.json')
            ppi_net = MagicMock()
            ppi_net.get_name = MagicMock(return_value='foo')
            ppi_net.to_cx = MagicMock(return_value={'foo': 'hi'})
            gen = CellmapsGenerateHierarchy(outdir='/foo')
            gen._write_ppi_network_as_cx(ppi_net, dest_path=afile)
            self.assertTrue(os.path.isfile(afile))
            with open(afile, 'r') as f:
                data = json.load(f)
            self.assertEqual({'foo': 'hi'}, data)
        finally:
            shutil.rmtree(temp_dir)

    def test_register_ppi_network(self):
        prov = MagicMock()
        prov.get_default_date_format_str = MagicMock(return_value='Y')
        ppi_net = MagicMock()
        ppi_net.get_name = MagicMock(return_value='net_name')
        prov.register_dataset = MagicMock(return_value='1')
        gen = CellmapsGenerateHierarchy(outdir='/foo',
                                        provenance_utils=prov)
        gen._description = 'description'
        gen._keywords = ['1', '1']
        res = gen._register_ppi_network(ppi_net, dest_path='/src/x')
        self.assertEqual(res, '1')
        self.assertEqual(('/foo',), prov.register_dataset.call_args.args)
        self.assertEqual('/src/x', prov.register_dataset.call_args.kwargs['source_file'])
        prov.register_dataset.assert_called_once()
        self.assertEqual(('/foo',), prov.register_dataset.call_args.args)
        self.assertEqual('/src/x', prov.register_dataset.call_args.kwargs['source_file'])
        d_dict_passed_in = prov.register_dataset.call_args.kwargs['data_dict']
        self.assertEqual('x PPI network file', d_dict_passed_in['name'])
        self.assertEqual('description PPI Network file', d_dict_passed_in['description'])
        self.assertEqual('CX', d_dict_passed_in['data-format'])
        self.assertEqual(cellmaps_generate_hierarchy.__name__, d_dict_passed_in['author'])
        self.assertEqual(cellmaps_generate_hierarchy.__version__, d_dict_passed_in['version'])
        self.assertEqual('Y', d_dict_passed_in['date-published'])
        self.assertTrue(2, len(d_dict_passed_in['keywords']))
        self.assertTrue('1' in d_dict_passed_in['keywords'])
        self.assertTrue('file' in d_dict_passed_in['keywords'])

    def test_register_ppi_network_keywords_is_none(self):
        prov = MagicMock()
        prov.get_default_date_format_str = MagicMock(return_value='Y')
        ppi_net = MagicMock()
        ppi_net.get_name = MagicMock(return_value='net_name')
        prov.register_dataset = MagicMock(return_value='1')
        gen = CellmapsGenerateHierarchy(outdir='/foo',
                                        provenance_utils=prov)
        gen._description = 'description'
        gen._keywords = None
        res = gen._register_ppi_network(ppi_net, dest_path='/src/x')
        self.assertEqual(res, '1')
        self.assertEqual(('/foo',), prov.register_dataset.call_args.args)
        self.assertEqual('/src/x', prov.register_dataset.call_args.kwargs['source_file'])
        prov.register_dataset.assert_called_once()
        self.assertEqual(('/foo',), prov.register_dataset.call_args.args)
        self.assertEqual('/src/x', prov.register_dataset.call_args.kwargs['source_file'])
        d_dict_passed_in = prov.register_dataset.call_args.kwargs['data_dict']
        self.assertEqual('x PPI network file', d_dict_passed_in['name'])
        self.assertEqual('description PPI Network file', d_dict_passed_in['description'])
        self.assertEqual('CX', d_dict_passed_in['data-format'])
        self.assertEqual(cellmaps_generate_hierarchy.__name__, d_dict_passed_in['author'])
        self.assertEqual(cellmaps_generate_hierarchy.__version__, d_dict_passed_in['version'])
        self.assertEqual('Y', d_dict_passed_in['date-published'])
        self.assertTrue(1, len(d_dict_passed_in['keywords']))
        self.assertTrue('file' in d_dict_passed_in['keywords'])

    def test_write_hierarchy_network(self):
        temp_dir = tempfile.mkdtemp()
        try:
            prov = MagicMock()
            prov.get_default_date_format_str = MagicMock(return_value='Y')
            ppi_net = MagicMock()
            ppi_net.get_name = MagicMock(return_value='net_name')
            prov.register_dataset = MagicMock(return_value='1')
            gen = CellmapsGenerateHierarchy(outdir=temp_dir,
                                            provenance_utils=prov)
            gen._description = 'description'
            gen._keywords = None
            res = gen._write_hierarchy_network(hierarchy=['hi'])
            self.assertEqual(os.path.join(temp_dir, 'hierarchy.cx2'), res)
            self.assertTrue(os.path.isfile(res))
            with open(res, 'r') as f:
                self.assertEqual(['hi'], json.load(f))
        finally:
            shutil.rmtree(temp_dir)

    # def test_register_hierarchy_network(self):





#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os
import re
from setuptools import setup, find_packages


with open(os.path.join('cellmaps_generate_hierarchy', '__init__.py')) as ver_file:
    for line in ver_file:
        line = line.rstrip()
        if line.startswith('__version__'):
            version = re.sub("'", "", line[line.index("'"):])
        elif line.startswith('__description__'):
            desc = re.sub("'", "", line[line.index("'"):])
        elif line.startswith('__repo_url__'):
            repo_url = re.sub("'", "", line[line.index("'"):])
        elif line.startswith('__author__'):
            author = re.sub("'", "", line[line.index("'"):])
        elif line.startswith('__email__'):
            email = re.sub("'", "", line[line.index("'"):])

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['cellmaps_utils',
                'requests',
                'tqdm',
                'pandas',
                'numpy',
                'ndex2>=3.5.1,<4.0.0',
                'cdapsutil>=0.2.2,<1.0.0',
                'leidenalg==0.9.1',
                'hidef>=1.1.5,<2.0.0']

setup_requirements = []

setup(
    author=author,
    author_email=email,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    description=desc,
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='cellmaps_generate_hierarchy',
    name='cellmaps_generate_hierarchy',
    packages=find_packages(include=['cellmaps_generate_hierarchy']),
    package_dir={'cellmaps_generate_hierarchy': 'cellmaps_generate_hierarchy'},
    package_data={'cellmaps_generate_hierarchy': ['interactome_style.cx2', 'interactome_style_with_bait.cx2',
                                                  'hierarchy_style.cx2', 'readme_outputs.txt']},
    scripts=['cellmaps_generate_hierarchy/cellmaps_generate_hierarchycmd.py'],
    url=repo_url,
    version=version,
    zip_safe=False)




import os
from datetime import date
import pandas as pd

import networkx as nx
import logging
import cellmaps_generate_hierarchy
from cellmaps_utils import constants
from cellmaps_utils.provenance import ProvenanceUtil

logger = logging.getLogger(__name__)


class HiDeFHierarchyRefiner(object):
    """
    Refines HiDeF hierarchy output by removing highly similar terms.
    This code derived from maturehierarchy.py developed by (## todo Leah who created this?)

    """

    TERMS_COL = 'terms'
    TSIZE_COL = 'tsize'
    GENES_COL = 'genes'
    STABILITY_COL = 'stability'

    TYPE_COL = 'type'
    PARENT_COL = 'parent'
    CHILD_COL = 'child'

    GENE_TYPE = 'gene'
    DEFAULT_TYPE = 'default'

    NODES_SUFFIX = '.nodes'
    EDGES_SUFFIX = '.edges'

    CONTAINMENT_THRESHOLD = 0.75
    JACCARD_THRESHOLD = 0.9
    MIN_DIFF = 1
    MIN_SYSTEM_SIZE = 4

    def __init__(self,
                 ci_thre=CONTAINMENT_THRESHOLD,
                 ji_thre=JACCARD_THRESHOLD,
                 min_term_size=MIN_SYSTEM_SIZE,
                 min_diff=MIN_DIFF,
                 provenance_utils=ProvenanceUtil(),
                 author='cellmaps_generate_hierarchy',
                 version=cellmaps_generate_hierarchy.__version__):
        """
        Constructor

        :param ci_thre: Containment index threshold
        :param ji_thre: Jaccard index threshold for merging similar clusters
        :param min_system_size: Minimum number of proteins requiring each system to have
        :param min_diff: Minimum difference in number of proteins for every parent-child pair
        """
        self._ci_thre = ci_thre
        self._ji_thre = ji_thre
        self._min_term_size = min_term_size
        self._min_diff = min_diff
        self._provenance_utils = provenance_utils
        self._author = author
        self._version = version

    @staticmethod
    def _get_node_table_from_hidef(nodes_file):
        """
        Takes a HiDeF nodes file and creates
        a :py:class:`pandas.DataFrame`

        **Example HiDeF edges file:**

        .. code-block::

            Cluster1-408    4       1528 2310 2666 4273     10
            Cluster1-409    4       126 3366 3794 446       10
            Cluster1-410    4       1522 2733 4334 4542     10

        **Resulting DataFrame:**

        .. code-block::

            terms           tsize   genes                   stability
            Cluster1-408    4       1528 2310 2666 4273     10
            Cluster1-409    4       126 3366 3794 446       10
            Cluster1-410    4       1522 2733 4334 4542     10

        :param nodes_file:
        :type nodes_file: str
        :return: DataFrame of the HiDeF nodes file
        :rtype: :py:class:`pandas.DataFrame`
        """
        node_table = pd.read_csv(nodes_file, header=None, sep='\t')
        logger.debug('nodes table head: ' + str(node_table.head()))
        logger.debug('Size of node table: ' + str(len(node_table)))

        node_table.columns = [HiDeFHierarchyRefiner.TERMS_COL,
                              HiDeFHierarchyRefiner.TSIZE_COL,
                              HiDeFHierarchyRefiner.GENES_COL,
                              HiDeFHierarchyRefiner.STABILITY_COL]
        return node_table

    @staticmethod
    def _get_edge_table_from_hidef(edges_file):
        """
        Takes a HiDeF edges file and creates
        a :py:class:`pandas.DataFrame`

        **Example HiDeF edges file:**

        .. code-block::

            Cluster0-0      Cluster1-0      default
            Cluster0-0      Cluster1-1      default
            Cluster0-0      Cluster1-2      default

        **Resulting DataFrame:**

        .. code-block::

            parent          child           type
            Cluster0-0      Cluster1-0      default
            Cluster0-0      Cluster1-1      default
            Cluster0-0      Cluster1-2      default


        :param edges_file:
        :type edges_file: str
        :return:
        :rtype: :py:class:`pandas.DataFrame`
        """
        edge_table = pd.read_csv(edges_file, header=None, sep='\t')
        edge_table.columns = [HiDeFHierarchyRefiner.PARENT_COL,
                              HiDeFHierarchyRefiner.CHILD_COL,
                              HiDeFHierarchyRefiner.TYPE_COL]
        return edge_table

    @staticmethod
    def _get_node_table_filtered_by_term_size(node_table, min_term_size=4):
        """
        Gets a new :py:class:`pandas.DataFrame` derived from
        **node_table** that only contains rows whose :py:const:`TSIZE_COL`
        is at least **min_term_size**

        :param min_term_size: Minimum size of term (ie number of genes)
        :type min_term_size: int
        :return: Filtered **node_table**
        :rtype: :py:class:`pandas.DataFrame`
        """
        filtered_table = node_table[node_table[HiDeFHierarchyRefiner.TSIZE_COL] >= min_term_size]
        logger.debug('Size node node table after term size filter: ' + str(len(filtered_table)))
        return filtered_table

    @staticmethod
    def _get_edge_table_filtered_by_node_set(edge_table, node_set=None):
        """
        Gets a :py:class:`pandas.DataFrame` derived from **edge_table** that only
        contains rows whose parent and child both exist in **node_set** passed in

        :param edge_table: table with :py:const:`PARENT_COL` and :py:const:`CHILD_COL`
                           columns
        :type edge_table: :py:class:`pandas.DataFrame`
        :param node_set: Names of clusters to keep
        :type node_set: set or list
        :return: edge table with rows whose parent and child both exist in **node_set** passed in
        :rtype: :py:class:`pandas.DataFrame`
        """
        return edge_table[edge_table[[HiDeFHierarchyRefiner.PARENT_COL,
                                      HiDeFHierarchyRefiner.CHILD_COL]].isin(list(node_set)).all(axis=1)]

    @staticmethod
    def _get_leaves_from_edge_table(edge_table=None):
        """
        Gets a set of term names found in :py:const:`CHILD_COL` column values
        that are **NOT** in :py:const:`PARENT_COL` column values

        :param edge_table:
        :type edge_table: :py:class:`pandas.DataFrame`
        :return: leave terms
        :rtype: set
        """
        return (set(edge_table[HiDeFHierarchyRefiner.CHILD_COL].unique())) - set(edge_table[HiDeFHierarchyRefiner.PARENT_COL].unique())

    def _get_leaf_genes_to_add(self, node_table=None, leaves=None):
        """
        For each term in **leaves** list get genes for that term and
        add a list with these elements ``[TERM, GENE NAME, GENE_TYPE]``
        to the list returned

        :param node_table:
        :param leaves: term name
        :type leaves: list
        :return: list of ``[TERM, GENE NAME, GENE_TYPE]`` for all the genes found
                in the terms in **leaves** list
        :rtype: list
        """
        add_gene = []
        for leaf in leaves:
            genes = self._get_genes_from_node_table_for_term(node_table, leaf)
            for gene in genes:
                add_gene.append([leaf, gene, HiDeFHierarchyRefiner.GENE_TYPE])
        return add_gene

    def _get_nonleaf_genes_to_add(self, node_table=None,
                                  edge_table=None):
        """
        # TODO: Test this
        Returns a list of ``[TERM, GENE NAME, GENE_TYPE]`` that
        correspond to genes **NOT** passed to leaf nodes

        :param node_table: HiDeF nodes table
        :type node_table: :py:class:`pandas.DataFrame`
        :param edge_table: HiDeF edge table
        :type edge_table: :py:class:`pandas.DataFrame`
        :return: list of ``[TERM, GENE NAME, GENE_TYPE]`` for all the genes found
                 in non-leaf nodes
        :rtype: list
        """
        gene_rows_to_add = []
        parent_to_child = edge_table.groupby(HiDeFHierarchyRefiner.PARENT_COL)[HiDeFHierarchyRefiner.CHILD_COL].apply(list)  # group the parents
        for parent, children in parent_to_child.items():
            parent_genes = self._get_genes_from_node_table_for_term(node_table, parent)
            child_genes = []
            for child in children:
                genes = self._get_genes_from_node_table_for_term(node_table, child)
                child_genes = set(child_genes).union(set(genes))
            # genes only in parent did not pass to child
            only_parent = set(parent_genes) - set(child_genes)
            if len(only_parent) >= 1:
                for gene in only_parent:
                    gene_rows_to_add.append([parent, gene, HiDeFHierarchyRefiner.GENE_TYPE])
        return gene_rows_to_add

    def _create_ontology(self, node_table=None, edge_table=None, min_term_size=4):
        """
        # TODO: Test this
        Creates ontology from HiDeF nodes and edges file

        :param path: Prefix path to nodes & edges files generated by HiDeF,
                     .nodes and .edges will be appended to this path
        :type path: str
        :param min_term_size:
        :type min_term_size: int
        :return:
        :rtype: py:class:`pandas.DataFrame`
        """
        # filter out rows below minTermSize
        node_table_filtered = self._get_node_table_filtered_by_term_size(node_table,
                                                                         min_term_size=min_term_size)

        logger.info(str(len(node_table_filtered)))
        node_set = set(node_table_filtered[HiDeFHierarchyRefiner.TERMS_COL])  # get the set of nodes

        # keep only edges that have entries in node_set
        edge_table_filtered = self._get_edge_table_filtered_by_node_set(edge_table, node_set=node_set)

        # Find leaves and get their genes
        leaves = self._get_leaves_from_edge_table(edge_table=edge_table_filtered)
        gene_rows_to_add = self._get_leaf_genes_to_add(node_table=node_table_filtered, leaves=leaves)

        # Get non leaf genes
        gene_rows_to_add.extend(self._get_nonleaf_genes_to_add(node_table=node_table_filtered,
                                                               edge_table=edge_table_filtered))

        # add gene rows to new dataframe and append to final dataframe
        add_rows = pd.DataFrame(gene_rows_to_add, columns=[HiDeFHierarchyRefiner.PARENT_COL,
                                                           HiDeFHierarchyRefiner.CHILD_COL,
                                                           HiDeFHierarchyRefiner.TYPE_COL])
        final_df = pd.concat([edge_table_filtered, add_rows])
        logger.debug(' Size of final table: ' + str(len(final_df)))
        return final_df

    @staticmethod
    def _get_genes_from_node_table_for_term(nodes_table, term):
        """

        Iterates over genes found under the :py:class:`GENES_COL` column
        splitting by space to get a gene list

        It is assumed the **nodes_table** has this format:

        .. code-block::

            terms           tsize   genes                   stability
            Cluster1-408    4       1528 2310 2666 4273     10
            Cluster1-409    4       126 3366 3794 446       10
            Cluster1-410    4       1522 2733 4334 4542     10


        :param nodes_table: Nodes table from HiDeF output
        :type nodes_table: :py:class:`pandas.DataFrame`
        :param term: term name used to find row to get genes from
        :type term: str
        :return: genes
        :rtype: list
        """
        genes = nodes_table.loc[nodes_table.terms == term][HiDeFHierarchyRefiner.GENES_COL]
        gene_list = []
        for g in genes:
            gene_list.extend(g.split(' '))
        return gene_list

    @staticmethod
    def _to_pandas_dataframe(nx_graph):
        """
        # TODO: Finish implementing this
        Converts **nx_graph** to a Pandas DataFrame
        with columns set to ``source, target, type``

        :param nx_graph:
        :return:
        :rtype: :py:class:`pandas.DataFrame`
        """
        e = nx_graph.edges(data=True)
        df = pd.DataFrame()
        df['source'] = [x[0] for x in e]
        df['target'] = [x[1] for x in e]
        df['type'] = [x[2]['type'] for x in e]
        return df

    @staticmethod
    def _get_term_stats(nx_graph, hiergeneset):
        # TODO: Finish implementing this
        clusters = list(set(list(nx_graph.nodes())) - hiergeneset)
        tsize_list = []
        cgene_list = []
        descendent_list = []
        for c in clusters:
            infoset = nx.descendants(nx_graph, c)
            cgeneset = infoset.intersection(hiergeneset)
            tsize_list.append(len(cgeneset))
            cgene_list.append(list(cgeneset))
            descendent_list.append(list(infoset - cgeneset))
        df = pd.DataFrame(index=clusters)
        df['tsize'] = tsize_list
        df['genes'] = cgene_list
        df['descendent'] = descendent_list
        return df

    @staticmethod
    def _jaccard(a, b):
        """
        Calculates Jaccard index of **a** and **b**

        :param a: set a
        :type a: list or set
        :param b: set b
        :type b: list or set
        :return: Jaccard index
        :rtype: float
        """
        if type(a) != set:
            a = set(a)
        if type(b) != set:
            b = set(b)
        return len(a.intersection(b)) / len(a.union(b))

    def _clean_shortcut(self, nx_graph):
        # TODO: Finish implementing this
        edge_df = self._to_pandas_dataframe(nx_graph)
        edge_df.columns = [HiDeFHierarchyRefiner.PARENT_COL,
                           HiDeFHierarchyRefiner.CHILD_COL,
                           HiDeFHierarchyRefiner.TYPE_COL]
        for idx, row in edge_df.iterrows():
            if len(list(nx.all_simple_paths(nx_graph, row[HiDeFHierarchyRefiner.PARENT_COL],
                                            row[HiDeFHierarchyRefiner.CHILD_COL]))) > 1:
                nx_graph.remove_edge(row[HiDeFHierarchyRefiner.PARENT_COL],
                                     row[HiDeFHierarchyRefiner.CHILD_COL])
                logger.debug('shortcut edges is removed between {} and {}'.format(row[HiDeFHierarchyRefiner.PARENT_COL],
                                                                                  row[HiDeFHierarchyRefiner.CHILD_COL]))

    def _reorganize(self, nx_graph, hiergeneset, ci_thre): # Add an edge if the nodes have containment index >=threshold
        # TODO: Finish implementing this
        iterate = True
        n_iter = 1
        while iterate:
            clear = True
            logger.debug('... starting iteration ' + str(n_iter))
            ts_df = self._get_term_stats(nx_graph, hiergeneset) # get the termStats from the networkx
            ts_df.sort_values('tsize', ascending=False, inplace=True)
            for comp, row in ts_df.iterrows():
                tmp = ts_df[ts_df['tsize'] < row['tsize']] # get all components smaller than this components
                if tmp.shape[0] == 0:
                    continue
                comp_geneset = set(row['genes']) # get the set of genes
                descendent = row['descendent'] # get the list of descendent nodes
                for tmp_comp, tmp_row in tmp.iterrows():
                    if tmp_comp in descendent: # skip if already in descendent
                        continue
                    tmp_comp_geneset = set(tmp_row['genes'])
                    # Check if satisfy ci_thre
                    # intersection of two components divided by the term size of the smaller component
                    if len(comp_geneset.intersection(tmp_comp_geneset))/tmp_row['tsize'] >= ci_thre:
                        # Check if child having higher weight than parent
                        logger.debug('{} is contained in {} with a CI bigger than threshold, add edge between'.format(tmp_comp, comp))
                        nx_graph.add_edge(comp, tmp_comp, type='default')
                        clear = False
                        descendent += tmp_row['descendent']
            # Further clean up using networkx to remove shortcut edges
            self._clean_shortcut(nx_graph)
            # Update variables
            n_iter += 1
            if clear:
                iterate = False
        if n_iter == 2:
            modified = False
        else:
            modified = True
        return modified

    def _merge_parent_child(self, nx_graph, hiergeneset, ji_thre):
        # TODO: Finish implementing this
        # Delete child term if highly similar with parent term
        # One parent-child relationship at a time to avoid complicacies involved in potential long tail
        logger.debug('... start removing highly similar parent-child relationship')
        similar = True
        merged = False
        while similar:
            clear = True
            edge_df = self._to_pandas_dataframe(nx_graph)
            ts_df = self._get_term_stats(nx_graph, hiergeneset)
            default_edge = edge_df[edge_df['type'] == 'default'] # edges
            for idx, row in default_edge.iterrows():
                if self._jaccard(ts_df.loc[row['source']]['genes'], ts_df.loc[row['target']]['genes']) >= ji_thre:
                    logger.debug('# Cluster pair {}->{} failed Jaccard, removing cluster {}'.format(row['source'], row['target'],
                                                                                             row['target']))
                    clear = False
                    merged = True
                    parents = edge_df[edge_df['target'] == row['target']]['source'].values
                    children = edge_df[edge_df['source'] == row['target']]['target'].values
                    # Remove all parent->node edges
                    for pnode in parents:
                        nx_graph.remove_edge(pnode, row['target'])
                    for child_node in children:
                        etype = nx_graph[row['target']][child_node]['type']
                        # Remove all node->child edges
                        nx_graph.remove_edge(row['target'], child_node)
                        # Add all parent->child edges
                        for pnode in parents:
                            nx_graph.add_edge(pnode, child_node, type=etype)
                    # Remove target node
                    nx_graph.remove_node(row['target'])
                    break
            if clear:
                similar = False
        # Clean up shortcuts introduced during node deleteing process
        self._clean_shortcut(nx_graph)
        return merged

    def _collapse_redundant(self, nx_graph, hiergeneset, min_diff):
        # TODO: Finish implementing this
        # Delete child term if highly similar with parent term
        # One parent-child relationship at a time to avoid complicacies involved in potential long tail
        logger.debug('... start removing highly redundant systems')
        while True:
            edge_df = self._to_pandas_dataframe(nx_graph)
            ts_df = self._get_term_stats(nx_graph, hiergeneset)
            default_edge = edge_df[edge_df['type'] == 'default']
            to_collapse = []
            for idx, row in default_edge.iterrows():
                parentSys, childSys, _ = row.values
                if ts_df.loc[parentSys]['tsize'] - ts_df.loc[childSys]['tsize'] < min_diff:
                    to_collapse.append([parentSys, childSys])
            if len(to_collapse) == 0:
                logger.debug('nothing to collapse')
                return
            to_collapse = pd.DataFrame(to_collapse, columns=['parent', 'child'])
            deleteSys = to_collapse.loc['child']
            logger.debug('# Cluster pair {}->{} highly redundant, removing cluster {}'.format(to_collapse.loc['parent'],
                                                                                       to_collapse.loc['child'],
                                                                                       deleteSys))
            parents = edge_df[edge_df['target'] == deleteSys]['source'].values
            children = edge_df[edge_df['source'] == deleteSys]['target'].values
            # Remove all parent->node edges
            for pnode in parents:
                nx_graph.remove_edge(pnode, deleteSys)
            for child_node in children:
                etype = nx_graph[deleteSys][child_node]['type']
                # Remove all node->child edges
                nx_graph.remove_edge(deleteSys, child_node)
                # Add all parent->child edges
                for pnode in parents:
                    nx_graph.add_edge(pnode, child_node, type=etype)
            # Remove target node
            nx_graph.remove_node(deleteSys)

    def _register_pruned_hidef_output_files(self, outprefix):
        """
        Register <outprefix>.nodes and <outprefix.edges> pruned
        HiDeF output files with FAIRSCAPE

        :param outprefix:
        :type outprefix: str
        :return: dataset ids
        :rtype: list
        """
        d_sets = []
        for hidef_file in ['nodes', 'edges']:
            outfile = outprefix + '.' + hidef_file
            data_dict = {'name': os.path.basename(outfile) +
                         ' Pruned HiDeF output ' + hidef_file[0] + ' file',
                         'description': ' Pruned HiDeF output ' + hidef_file[0] + ' file',
                         'data-format': 'tsv',
                         'author': str(self._author),
                         'version': str(self._version),
                         'date-published': date.today().strftime('%m-%d-%Y')}
            d_sets.append(self._provenance_utils.register_dataset(os.path.dirname(outfile),
                                                                  source_file=outfile,
                                                                  data_dict=data_dict))
        return d_sets

    def refine_hierarchy(self, outprefix=None):
        """
        Removes highly similar systems and dumps out a new HiDeF formatted
        .nodes and .edges file with .pruned.nodes and .pruned.edges suffixes

        :param outprefix: output_dir/file_prefix for the output file
        :type outprefix: str
        :return: dataset ids of .pruned.nodes and .pruned.edges file generated
        :rtype: list
        """
        logger.debug('Containment index threshold: ' + str(self._ci_thre))
        logger.debug('Jaccard index threshold: ' + str(self._ji_thre))

        # Read node table
        node_table = self._get_node_table_from_hidef(outprefix +
                                                     HiDeFHierarchyRefiner.NODES_SUFFIX)

        # Read edge table
        edge_table = self._get_edge_table_from_hidef(outprefix + HiDeFHierarchyRefiner.EDGES_SUFFIX)

        ont = self._create_ontology(node_table=node_table, edge_table=edge_table,
                                    min_term_size=self._min_term_size)
        hiergeneset = set(ont[ont[HiDeFHierarchyRefiner.TYPE_COL] == HiDeFHierarchyRefiner.GENE_TYPE][HiDeFHierarchyRefiner.CHILD_COL].values)

        nx_graph = nx.from_pandas_edgelist(ont,
                                           source=HiDeFHierarchyRefiner.PARENT_COL,
                                           target=HiDeFHierarchyRefiner.CHILD_COL,
                                           edge_attr=HiDeFHierarchyRefiner.TYPE_COL,
                                           create_using=nx.DiGraph())

        if not nx.is_directed_acyclic_graph(nx_graph):
            raise ValueError('Input hierarchy is not DAG!')

        while True:
            modified = self._reorganize(nx_graph, hiergeneset, self._ci_thre)
            merged = self._merge_parent_child(nx_graph, hiergeneset, self._ji_thre)
            if not modified and not merged:
                break

        self._collapse_redundant(nx_graph, hiergeneset, self._min_diff)
        # Output as ddot edge file
        self._clean_shortcut(nx_graph)

        edge_df = self._to_pandas_dataframe(nx_graph)

        # we need to recreate .nodes file in HiDeF format
        # so create nodes dataframe which has this mapping of
        # all genes for a given cluster (including genes attached to children)
        nodes = self._get_term_stats(nx_graph, hiergeneset)
        logger.debug(nodes.head())

        # get rid of the descendent column
        nodes.drop(['descendent'], axis=1, inplace=True)

        # the genes are in a list, but we need them space delimited so
        # in place replace the 'genes' column
        cleaned = []
        for i, rows in nodes.iterrows():
            if isinstance(rows['genes'], list):
                genes = rows['genes']
            else:
                genes = [g for g in rows['genes'].split(',') if len(g) > 1]
            cleaned.append(' '.join(genes))
        nodes['genes'] = cleaned

        # dont think this is needed
        # node_table.set_index(HiDeFHierarchyRefiner.TERMS_COL)

        # create a map of cluster name to stability values
        cluster_stability = pd.Series(node_table[HiDeFHierarchyRefiner.STABILITY_COL].values,
                                      index=node_table.terms).to_dict()

        # use the cluster_stability map to add stability values to nodes dataframe by
        # matching with the cluster name in the nodes (it is the index)
        nodes[HiDeFHierarchyRefiner.STABILITY_COL] = nodes.index.map(cluster_stability)

        # sort the nodes table by term size
        nodes.sort_values(by='tsize', ascending=False, inplace=True)

        # Todo: Register .pruned.nodes and .pruned.edges with fairscape

        # dump out new nodes file in HiDeF format
        nodes.to_csv(outprefix + '.pruned.nodes', header=False, sep='\t')

        # dump out HiDeF edges file
        edges = edge_df.loc[edge_df['type'] == 'default', :]
        edges.to_csv(outprefix+'.pruned.edges', sep='\t', header=None,index=None)

        logger.debug('Number of edges is ' + str(len(edges)) + ', number of nodes are ' + str(len(nodes)))
        if self._provenance_utils is not None:
            return self._register_pruned_hidef_output_files(outprefix + '.pruned')
        else:
            return list()
